from datetime import datetime
import re

SYS_POS = slice(5, 6)
SYS_EXT_POS = slice(5, 7)
RUN_POS = slice(7, 9)
DEV_START_POS = 9  # slice(9, 12)
VSN_POS = slice(1, 3)


class psense_exp_title:
    def __init__(self):
        # process experiment Information
        self.typeof_study = dict(
            {
                "H": "Clinical",
                "F": "Flow System",
                "B": "Beaker",
                "A": "Heat Block",
                "P": "Prototype",
                "R": "Pre-Clinical",
            }
        )

        self.typeof_sensor = dict(
            {
                "G": "Glucose",
                "O": "Oxygen",
                "L": "Lactate",
                "K": "Ketone",
                "T": "Temperature",
                "X": "Multi (legacy)",
                "A": "Acetylcholine",
                "C": "Choline",
                "N": "Other",
            }
        )
        self.typeof_electrode = dict(
            {"C": "Circular", "R": "Rectangular", "X": "Other"}
        )
        self.typeof_stack = dict(
            {
                "R": "Ring",
                "P": "Planar",
                "X": "Other",
            }
        )
        self.typeof_process = dict(
            {
                "S": "Standard",
                "V": "Volcano",
                "B": "Blanket",
                "W": "Sandwich",
                "0": "No Silicone",
            }
        )
        self.typeof_extras = dict(
            {
                "M": "Mediator",
                "C": "Catalase",
                "P": "Peroxidase",
                "N": "NAD+",
                "O": "Other",
                "X": "X",
            }
        )

    def decode(self, expid):
        """translate a percusense experiment identifier string into a human-
        readble dictionary"""

        res = []
        try:
            regions = expid.upper().split("-")

            # should have 5 "regions" in an experiment code -- attempt to handle partial codes
            regions.extend([""] * (5 - len(regions)))

            if len(regions[0]) > 4:
                study = {
                    "start": datetime.strptime(
                        "{}-{}-{}".format(
                            regions[0][:2],
                            max(ord(regions[0][2:3]) - 64, 1),
                            max(1, int(regions[0][3:5])),
                        ),
                        "%y-%m-%d",
                    )
                    .date()
                    .isoformat(),
                    "type": self.typeof_study.get(regions[0][SYS_POS]),
                    "run": regions[0][RUN_POS],
                    "device": regions[0][DEV_START_POS:],
                }

                # special cases -- 2-digit study type
                if regions[0][SYS_EXT_POS] == "AN":
                    study["type"] = "Pre-Clinical"
                elif regions[0][SYS_EXT_POS] == "HU":
                    study["type"] = "Clinical"

            else:
                study = []

            if len(regions[2]) > 0:
                sensor = dict()
                sensor["code"] = "{}-{}-{}".format(regions[1], regions[2], regions[3])
                sensor["count"] = regions[4][0]
                sensor["vsn"] = regions[4][VSN_POS]

                electrodes = []
                if sensor["count"].isalpha():
                    # legacy, or single-analyte
                    text = regions[4]
                    electrodes = [
                        dict(
                            type=self.typeof_sensor.get(text[0]),
                            electrode=self.typeof_electrode.get(text[3]),
                            stack=self.typeof_stack.get(text[4]),
                            process=self.typeof_process.get(text[5]),
                            extras=self.typeof_extras.get(text[6]),
                        )
                    ]
                else:
                    # multi-analyte
                    for elec in range(int(regions[4][:1])):
                        offset = 3 + (elec * 3)
                        text = regions[4][slice(offset, offset + 3)]
                        res = dict(
                            type=self.typeof_sensor.get(text[0]),
                            electrode=self.typeof_electrode.get(text[1]),
                            process=self.typeof_process.get(text[2]),
                        )
                        electrodes.append(res)

                sensor["electrodes"] = electrodes
            else:
                sensor = []

            res = dict(study=study, sensor=sensor)
        except BaseException:
            pass

        return res


"Simple IO Functions"


def yn_input(message, default="y"):
    "i/o: collect a yes/no answer from input"
    choices = "Y/n" if default.lower() in ("y", "yes") else "y/N"
    choice = input("{} ({}) ".format(message, choices))
    values = ("y", "yes", "") if choices == "Y/n" else ("y", "yes")
    return choice.strip().lower() in values


def answer_input(message, choices, default="0", is_required=True):
    "i/o: collect a typed answer from input (regex response validation)"
    is_valid = False

    if isinstance(choices, list):
        while not is_valid:
            choice_str = ", ".join(choices)
            choice = input("{} ({})".format(message, choice_str)).strip().upper()
            if len(choice) == 0 and not is_required:
                return default
            elif len(choice) > 0 and choice in choices:
                return choice

            print("Invalid Input ({})".format(choice))

    elif isinstance(choices, str):
        while not is_valid:
            choice = input(message).strip().upper()
            if len(choice) == 0 and not is_required:
                return default
            elif len(choice) > 0 and choice in str(
                re.search(r"{}".format(choices.upper()), choice)
            ):

                return choice

            print("Invalid Input ({})".format(choice))

    assert False


def timestamp_input(message, use_seconds=False, is_required=True):
    is_valid = False
    while not is_valid:
        entry = (
            input(
                "{} (yyyy-mm-dd HH:MM{}): ".format(
                    message, ":SS" if use_seconds else ""
                )
            )
            .strip()
            .upper()
        )
        try:
            entry = datetime.strptime(
                entry, "%Y-%m-%d %H:%M{}".format(":%S" if use_seconds else "")
            )
            is_valid = True
        except ValueError:
            print(" > invalid entry.")
            pass

    return entry


def exptype_input():
    "i/o: selection of starting a new experiment or resuming an existing one"
    message = "\nSelect experiment type and press enter\n\t[1] Start a new sensor (default)\n\t[2] Resume an existing sensor\n > "
    isvalid = False
    while not isvalid:
        choice = input(message).strip()
        if choice == "":
            return True
        elif choice.isnumeric():
            choice = int(choice)
            if choice == 2:
                return False
            else:
                return True
        else:
            print("Invalid input.")


def setup_new_experiment(device):
    "i/o: user input to generate a new percusense experiment id (new format as of 07/23/2020)"
    assert isinstance(device, str), "device must be of type string"

    exp_beaker = answer_input(
        "Specify the test system:\n\tA#: Heat Block\n\tB#: Beaker\n\tF#: Flow System\n\tH#: Clinical (also, HU)\n\tR#: Pre-Clinical (also, AN)\n",
        "([ABFPHR][0-9]|AN|HU){1}",
        "ZZ",
        True,
    )
    exp_run = int(
        answer_input("Specify the run number [1]: ", "[0-9]{1,2}", "1", False)
    )
    exp_lot_we = answer_input(
        "Mfg Identifier 1 (Lot, required): ", "[0-9]{1,3}[A-Z0-9]{0,1}", "1", True
    )
    exp_lot_re = int(answer_input("Mfg Identifier 2 [000]: ", "[0-9]{0,3}", "0", False))
    exp_lot_sn = answer_input("Sensor ID [00]: ", "[A-Z0-9]{0,2}", "00", False)

    exp_vsn = int(answer_input("VSN Substrate Design [15]: ", "[0-9]{1,2}", "15", True))

    exp_we_count = int(
        answer_input("How many WE are connected? [2]: ", "[0-9]{1}", "2", False)
    )

    electrodes = []
    for electrode in range(exp_we_count):
        print("> Enter Information about WE #{}".format(electrode + 1))
        exp_type = answer_input(
            "Sensor Type ([G]lucose, (O)xygen, (L)actate, (K)etone, (T)emp, (A)cetylcholine, (C)holine, (N) Other): ",
            "[GOLKACNT]",
            "G",
            False,
        )
        exp_discs = (
            answer_input(
                "Electrode Shape ([C]ircular, (R)ectangular, (X)): ",
                "[A-Z]{1}",
                "C",
                False,
            )
            .strip()
            .upper()
        )
        exp_volcano = (
            answer_input(
                "Process Label ([V]olcano, (S)tandard, (B)lanket, Sand(W)ich, (X)): ",
                "[A-Z]{1}",
                "V",
                True,
            )
            .strip()
            .upper()
        )
        electrodes.append(dict(type=exp_type, electrode=exp_discs, process=exp_volcano))

    region1 = "{}{}{}{}{:02d}{:0>3.3}".format(
        datetime.now().strftime("%y"),
        chr(int(datetime.now().strftime("%m")) + 64),
        datetime.now().strftime("%d"),
        exp_beaker,
        exp_run,
        device,
    )
    region2 = "{:0>4.4}-{:03d}-{:.2}".format(
        exp_lot_we,
        exp_lot_re,
        exp_lot_sn.zfill(2),
    )
    region3 = "{}{:02d}{}".format(
        exp_we_count,
        exp_vsn,
        "".join(
            [
                "{}{}{}".format(elec["type"], elec["electrode"], elec["process"])
                for elec in electrodes
            ]
        ),
    )
    expid = "{}-{}-{}".format(region1, region2, region3).strip().upper()

    print("> done with data input. ")
    print("Experiment ID: {}".format(expid))

    return expid



import numpy as np
import re


class kfilter(object):
    "PercuSense filter generates kalman and savitzky-golay outputs that are used for data smoothing/estimation"

    def __init__(
        self,
        filter_length=10,
        filter_init=40,
        process_variance=0.01,
        initial_guess=5,
        measurements=[],
        error_guess=0.01,
        rep_count=0,
    ):
        self.debugmode = False
        self.filter_length = filter_length

        # disable filter for first 40 points (~20min)
        self.filter_init = filter_init
        # 5% change allowed between consecutive measurements
        self.clip_pct_high = 1.05
        self.clip_pct_low = 0.95

        # kalman filter coefficients
        self.process_variance = process_variance
        self.posteri_estimate = initial_guess
        self.posteri_error_estimate = error_guess

        # filter initialization + state
        self.measurements = measurements
        self.consecutive_replacements = rep_count
        self.posteri_noise = 0
        self.posteri_deriv = 0

        # savitzky-golay filter first derivative coeff -- use this for rate term
        self.coeff_sg_velocity = [-3, -2, -1, 0, 3, 1, 2]
        self.denom_sg_velocty = 28
        # savitzky-golay filter second derivative coeff -- accel + noise calcs
        self.coeff_sg_accel = [5, 0, -3, -4, -3, 0, 5]
        self.denom_sg_accel = 42

    def in_init(self):
        return self.filter_init > 0

    def calculate_derivative(self, measurements=None):
        if measurements is None:
            measurements = self.measurements
            """if self.in_init():
                return 0"""

        # savitzky-golay first derivative filter
        coeff = self.coeff_sg_velocity
        normalization = self.denom_sg_velocty

        if len(measurements) < 7:
            return 0
        else:
            dp = np.dot(coeff, measurements[-7:])
            return dp / normalization

    def calculate_accel(self, measurements=None):
        if measurements is None:
            measurements = self.measurements
            """if self.in_init():
                return 0"""

        # savitzky-golay second derivative filter
        coeff = self.coeff_sg_accel
        normalization = self.denom_sg_accel

        if len(measurements) < 7:
            return 0
        else:
            dp = np.dot(coeff, measurements[-7:])
            return dp / normalization

    def update_measurement_state(self, measurement_value):
        "update kalman states"
        if (
            # filter fully populated
            (len(self.measurements) >= (self.filter_length))
            # filter doesn't need to be reset
            and (self.consecutive_replacements < (self.filter_length))
            and (not self.in_init())  # no during init
        ):
            latest_measurement = self.measurements[-1]

            # simple validation -- don't alow raw signal to change to 3/4 or 4/3 latestbetween samples
            if np.mean(self.measurements) >= 0:
                # positive signals
                if measurement_value > self.clip_pct_high * latest_measurement:
                    measurement_value = self.clip_pct_high * latest_measurement
                    self.consecutive_replacements += 1
                elif measurement_value < self.clip_pct_low * latest_measurement:
                    measurement_value = self.clip_pct_low * latest_measurement
                    self.consecutive_replacements += 1
                else:
                    self.consecutive_replacements = 0
            else:
                # negative signals
                if measurement_value < self.clip_pct_high * latest_measurement:
                    measurement_value = self.clip_pct_high * latest_measurement
                    self.consecutive_replacements += 1
                elif measurement_value > self.clip_pct_low * latest_measurement:
                    measurement_value = self.clip_pct_low * latest_measurement
                    self.consecutive_replacements += 1
                else:
                    self.consecutive_replacements = 0

            self.measurements.append(round(measurement_value, 2))

            # if the filter has been fully populated
            if len(self.measurements) > self.filter_length:
                self.measurements.pop(0)

            # noise = measurement sigma over the last n-samples
            self.posteri_noise = round(np.std(self.measurements), 3)

            # update the rate
            self.posteri_deriv = round(self.measurements[-1] - self.posteri_estimate, 3)

            """
            print('{},{:.3},{:.3}'.format(self.posteri_estimate, self.posteri_deriv, round(
                np.mean([t - s for s, t in zip(self.measurements, self.measurements[1:])][-3:]), 3)))

            # self.posteri_deriv = round(
            #    np.mean([t - s for s, t in zip(self.measurements, self.measurements[1:])][-3:]), 3)
            if self.filter_length < 7:
                # this is a simple calculation. Difference between previous estimate and new measurement.
                self.posteri_deriv = round(
                    self.measurements[-1] - self.posteri_estimate, 3)
            else:
                self.posteri_deriv = self.calculate_derivative(
                    self.measurements)
            """
        else:
            if self.consecutive_replacements > 0:
                self.consecutive_replacements = 0
                self.measurements = self.measurements[
                    min(round(self.filter_length / 2), len(self.measurements)) :
                ]

            self.measurements.append(round(measurement_value, 2))
            self.posteri_estimate = measurement_value

            # if the filter has been fully populated
            if len(self.measurements) > self.filter_length:
                self.measurements.pop(0)

            # if we are in init, decrement the # of samples remaining for init mode
            if self.in_init():
                self.filter_init -= 1

            # don't update noise, don't update rate of change
            pass

    def input_latest_measurement(self, measurement_value):
        "update the filter with new raw data"
        if isinstance(measurement_value, str):
            measurement_value = float(measurement_value)

        assert isinstance(measurement_value, (int, float))

        # update posteri_noise and posteri_deriv values
        self.update_measurement_state(measurement_value)
        priori_estimate = self.posteri_estimate
        priori_error_estimate = self.posteri_error_estimate + self.process_variance

        blending_factor = priori_error_estimate / (
            priori_error_estimate + self.posteri_noise
        )

        self.posteri_estimate = round(
            priori_estimate + blending_factor * self.posteri_deriv, 3
        )
        self.posteri_error_estimate = round(
            (1 - blending_factor) * priori_error_estimate, 4
        )

    def get_latest_estimated_measurement(self):
        "return kalman estimate along with the clipped input stored in the measurement buffer"
        return self.posteri_estimate, self.measurements[-1]

    def run(self, data):
        "generator for retrieving full filter outputs"

        for val in data:
            self.input_latest_measurement(val)
            filt, clip = self.get_latest_estimated_measurement()
            velocity = self.calculate_derivative()
            acceleration = self.calculate_accel()
            init = self.in_init()
            yield init, clip, filt, velocity, acceleration


def blend(signal1, signal2, noise, blending_factor=0.05):
    "blends two signals based on a normalizing noise factor"
    return ((signal1 * blending_factor) + (signal2 * noise)) / (noise + blending_factor)


def run_psensefilter(
    data,
    blending_factor=0.05,
    samples_mavg=40,
    samples_stdev=10,
    kalman_coeff={"filter-window": 30, "filter-smoothing": 0.01, "filter-init": 40},
):
    "apply kfilter and blending to pandas dataframe (raw data in col signal#)"
    data = data.copy(deep=True)  # only change the data within this scope
    keys = data.columns.values.tolist()
    r = re.compile("signal*")

    for channel in list(filter(r.match, keys)):
        signal = data[channel]
        kf = kfilter(
            filter_length=kalman_coeff["filter-window"],
            filter_init=kalman_coeff["filter-init"],
            process_variance=kalman_coeff["filter-smoothing"],
            initial_guess=signal[0],
        )

        clip = []
        fsig = []
        dsig = []
        ddsig = []
        smooth = []
        noise = []
        for init, a, b, c, d in kf.run(signal):
            clip.append(a)
            fsig.append(b)
            dsig.append(c)
            ddsig.append(d)

            if init or len(clip) < samples_mavg:
                smooth.append(np.mean(clip))
                noise.append(0)
            else:
                smooth.append(np.mean(clip[(1 - samples_mavg) :]))
                noise.append(np.std(ddsig[(1 - samples_stdev) :]))

        del kf  # not sure why but kf isn't re-initializing for new channels

        data["c_{}".format(channel)] = clip
        data["fk_{}".format(channel)] = fsig
        data["f{}_{}".format(samples_mavg, channel)] = smooth = np.array(smooth)
        data["d_{}".format(channel)] = dsig
        data["dd_{}".format(channel)] = ddsig
        data["noise{}_{}".format(samples_stdev, channel)] = noise = np.array(noise)
        data["fblend_{}".format(channel)] = blend(
            np.add(fsig, dsig), smooth, noise, blending_factor
        )

    data = data.round(2)
    return data



import unittest
import pandas as pd
import numpy as np
from psense_common.analysis import PSenseCalibration
from datetime import datetime


class TestPCal(unittest.TestCase):
    def setUp(self):
        self.test_expid = "19D01B101G03-0229-229-HT-G11CPBX"
        self.test_sensor = pd.read_csv(
            "tests/testdata_calibration_sensor",
            sep=",",
            index_col="timestamp",
            parse_dates=["timestamp"],
        )

    def test_init(self):
        cal = PSenseCalibration(expid=self.test_expid, data=self.test_sensor)

        self.assertEqual(cal.expid, self.test_expid)
        self.assertEqual(list(cal.result.keys()), ["coeff", "r"])
        self.assertEqual(cal.channels, ["signal1"])
        self.assertEqual(len(cal.raw), 319)

    def test_get_dataslice(self):
        cal = PSenseCalibration(expid=self.test_expid, data=self.test_sensor)

        t_start = pd.Timestamp("2019-04-02 06:30:00")
        t_end = pd.Timestamp("2019-04-02 06:40:00")
        res = cal.get_sensordata_slice(t_start, t_end)
        self.assertEqual(len(res), 20)
        self.assertEqual(res.iloc[0]["signal1"], 2.72)
        self.assertEqual(res.iloc[-1]["signal1"], 2.73)
        self.assertEqual(res.index[0], pd.Timestamp("2019-04-02 06:30:10"))

        t_end = t_start
        t_start = pd.Timestamp("2019-04-01 06:30:00")
        res = cal.get_sensordata_slice(t_start, t_end)
        self.assertEqual(len(res), 0)

        t_start = pd.Timestamp("2019-04-02 10:55:00")
        t_end = pd.Timestamp("2019-04-03 01:00:00")
        res = cal.get_sensordata_slice(t_start, t_end)
        self.assertEqual(len(res), 10)
        self.assertEqual(res.index[-1], pd.Timestamp("2019-04-02 10:59:49"))

    def test_add_reference(self):
        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )

        # confirm a point can be added
        ts = pd.Timestamp("2019-04-02 08:55:00")
        cal.add_reference(ts, 100)
        self.assertEqual(cal.calibration.iloc[0]["signal1"], 2.73)
        self.assertEqual(cal.calibration.iloc[0]["ref"], 100)
        self.assertEqual(
            cal.calibration.iloc[0]["ts"], pd.Timestamp("2019-04-02 08:55:00")
        )

        # confirm offset and window will change based on initialization
        cal = PSenseCalibration(
            expid=self.test_expid,
            data=self.test_sensor,
            sample_offset_time=0,
            sample_window_time=10,
        )
        ts = pd.Timestamp("2019-04-02 08:55:00")
        cal.add_reference(ts, 200)
        self.assertEqual(cal.calibration.iloc[0]["signal1"], 2.82)
        self.assertEqual(cal.calibration.iloc[0]["ref"], 200)

        # should add null value when sensor data doesn't exist
        ts = pd.Timestamp("2019-04-03 08:55:00")
        cal.add_reference(ts, 100)
        self.assertTrue(pd.isnull(cal.calibration.iloc[-1]["signal1"]))
        self.assertEqual(len(cal.calibration), 2)

    def test_linear_reg(self):
        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )
        self.assertRaises(AssertionError, cal.model_fit)

        cal.add_reference(pd.Timestamp("2019-04-02 09:00:00"), 100)
        cal.add_reference(pd.Timestamp("2019-04-02 09:30:00"), 200)
        cal.add_reference(pd.Timestamp("2019-04-02 10:00:00"), 300)
        cal.add_reference(pd.Timestamp("2019-04-02 10:30:00"), 400)

        cal.model_fit()
        self.assertTrue(len(cal.result["coeff"]) == 1)
        self.assertEqual(cal.result["coeff"][0], [0.0208, 0.62])
        self.assertEqual(cal.result["r"][0], [0.9986])

        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )
        cal.add_reference(pd.Timestamp("2019-04-02 09:00:00"), 400)
        cal.add_reference(pd.Timestamp("2019-04-02 09:30:00"), 300)
        cal.add_reference(pd.Timestamp("2019-04-02 10:00:00"), 200)
        cal.add_reference(pd.Timestamp("2019-04-02 10:30:00"), 100)
        cal.model_fit()
        self.assertEqual(cal.result["coeff"][0], [-0.0208, 11.0])
        self.assertEqual(cal.result["r"][0], [-0.9986])

    def test_logarithmic_reg(self):
        self.test_sensor = pd.read_csv(
            "tests/testdata_calibration_logfit",
            sep=",",
            index_col="timestamp",
            parse_dates=["timestamp"],
        )
        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )
        cal.add_reference(pd.Timestamp("2019-04-02 09:00:00"), 0)
        cal.add_reference(pd.Timestamp("2019-04-02 09:30:00"), 1)
        cal.add_reference(pd.Timestamp("2019-04-02 10:00:00"), 2)
        cal.add_reference(pd.Timestamp("2019-04-02 10:30:00"), 3)

        cal.model_fit()
        self.assertEqual(cal.result["r"][0], 0.9418)
        cal.result = dict(coeff=[], r=[])

        cal.model_fit(model_type="logarithmic")
        self.assertTrue(
            np.array_equal(
                cal.result["coeff"][0],
                np.array([-0.668813166669576, 0.8773716434543674, 1.0801832613942033]),
            )
        )
        self.assertEqual(cal.result["r"][0], 0.9993)

    def test_multichan(self):
        self.test_sensor = pd.read_csv(
            "tests/testdata_calibration_multichan",
            sep=",",
            index_col="ts",
            parse_dates=["ts"],
        )
        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )
        self.assertEqual(cal.channels, ["signal1", "signal2"])

        cal.add_reference(pd.Timestamp("2019-04-02 8:30"), 100)
        cal.add_reference(pd.Timestamp("2019-04-02 9:00"), 200)
        cal.add_reference(pd.Timestamp("2019-04-02 9:45"), 300)
        cal.model_fit()
        self.assertEqual(cal.result["coeff"][0], [0.0106, 0.57])
        self.assertEqual(cal.result["coeff"][1], [0.0107, 0.7])
        self.assertEqual(cal.result["r"], [0.9909, 0.9940])

    def test_dump_creates_deepcopy(self):
        cal = PSenseCalibration(
            expid=self.test_expid, data=self.test_sensor, debugmode=False
        )
        cal.add_reference(pd.Timestamp("2019-04-02 09:00:00"), 100)
        mycopy = cal.dump()
        cal.add_reference(pd.Timestamp("2019-04-02 09:30:00"), 200)

        self.assertEqual(len(mycopy), 1)
        self.assertEqual(len(cal.calibration), 2)


if __name__ == "__main__":
    unittest.main()



import unittest
import pandas as pd
from psense_common import kfilter
from datetime import datetime, timedelta


class TestKFilter(unittest.TestCase):
    def test_invalid_input(self):
        print("\ntest_invalid_input:")
        print("initialize class using fully populated buffer")
        kf = kfilter.kfilter()

        print("> should not allow inputting string")
        # kf.input_latest_measurement('teststring')
        self.assertRaises(
            AssertionError, kf.input_latest_measurement, ["StringInsteadOfNumber"]
        )
        self.assertEqual(kf.posteri_estimate, 5)
        self.assertEqual(kf.posteri_error_estimate, 0.01)
        self.assertEqual(kf.measurements, [])
        self.assertEqual(kf.consecutive_replacements, 0)

        print("> should not allow inputting list")
        self.assertRaises(AssertionError, kf.input_latest_measurement, [[10, 10]])

        print("> should allow inputting numeric string")
        kf.input_latest_measurement("10.10")
        self.assertEqual(kf.posteri_estimate, 10.1)
        self.assertEqual(kf.measurements, [10.10])

    def test_initial_state(self):
        print("\ntest_initial_state:")
        print("> test initalization with default values")
        kf = kfilter.kfilter()
        self.assertEqual(kf.filter_length, 10)
        self.assertEqual(kf.process_variance, 0.01)
        self.assertEqual(kf.posteri_estimate, 5)
        self.assertEqual(kf.posteri_error_estimate, 0.01)
        self.assertEqual(kf.measurements, [])
        self.assertEqual(kf.consecutive_replacements, 0)

        print("> test initalization with custom values")
        kf = kfilter.kfilter(1, 0, 1, 1, [1], 1, 1)
        self.assertEqual(kf.filter_length, 1)
        self.assertEqual(kf.filter_init, 0)
        self.assertEqual(kf.process_variance, 1)
        self.assertEqual(kf.posteri_estimate, 1)
        self.assertEqual(kf.posteri_error_estimate, 1)
        self.assertEqual(kf.measurements, [1])
        self.assertEqual(kf.consecutive_replacements, 1)

    def test_add_meas_full_buffer(self):
        print("\ntest_add_meas_full_buffer:")
        print("initialize class using fully populated buffer")
        kf = kfilter.kfilter(
            filter_length=3, filter_init=0, initial_guess=10, measurements=[10, 10, 10]
        )

        # print('{:06.3f}\t{:07.5f}\t{}\t{}'.format(kf.posteri_estimate, kf.posteri_error_estimate, kf.consecutive_replacements, kf.measurements))
        print("> test adding higher measurement, in range")
        kf.input_latest_measurement(10.5)
        self.assertEqual(kf.measurements, [10, 10, 10.5])
        self.assertEqual(kf.get_latest_estimated_measurement(), (10.039, 10.5))
        self.assertEqual(round(kf.posteri_error_estimate, 3), 0.018)

        print("> test adding higher measurement, clipping filter")
        kf.input_latest_measurement(10)
        self.assertEqual(kf.get_latest_estimated_measurement(), (10.035, 10))
        kf.input_latest_measurement(10)
        self.assertEqual(kf.get_latest_estimated_measurement(), (10.03, 10))
        kf.input_latest_measurement(10)
        self.assertEqual(kf.get_latest_estimated_measurement(), (10, 10))
        kf.input_latest_measurement(12)
        self.assertEqual(kf.measurements, [10, 10, 10.5])

        print("> test adding higher measurement, not clipped but out of range")
        self.assertEqual(kf.get_latest_estimated_measurement(), (10.02, 10.5))
        self.assertEqual(round(kf.posteri_error_estimate, 3), 0.010)

        print("> test adding higher measurement, edge of range")
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10.5)
        self.assertEqual(kf.measurements, [10, 10, 10.5])

        print("> test adding lower measurement, out of range")
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(5.0)
        self.assertEqual(kf.measurements, [10, 10, 9.5])
        self.assertEqual(kf.get_latest_estimated_measurement(), (09.98, 9.5))

        print("> test adding lower measurement, edge of range")
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(10)
        kf.input_latest_measurement(9.5)
        self.assertEqual(kf.measurements, [10, 10, 9.5])
        self.assertEqual(kf.get_latest_estimated_measurement(), (09.956, 9.5))

        print("initialize class using fully populated buffer, negative current")
        kf = kfilter.kfilter(
            filter_length=3, filter_init=0, initial_guess=-9, measurements=[-9, -9, -9]
        )

        print("> test adding more negative measurement, in range")
        kf.input_latest_measurement(-9.1)
        self.assertEqual(kf.measurements, [-9, -9, -9.1])
        self.assertEqual(kf.get_latest_estimated_measurement(), (-9.03, -9.1))
        self.assertEqual(round(kf.posteri_error_estimate, 3), 0.014)

        print("> test adding more negative measurement, clipped")
        kf.input_latest_measurement(-9)
        kf.input_latest_measurement(-9)
        kf.input_latest_measurement(-9)
        kf.input_latest_measurement(-10)
        self.assertEqual(kf.measurements, [-9, -9, -9.45])
        self.assertEqual(kf.get_latest_estimated_measurement(), (-9.02, -9.45))
        self.assertEqual(round(kf.posteri_error_estimate, 3), 0.010)

    def test_buffer_replacements(self):
        print("\ntest_buffer_replacements:")
        print("initialize class using fully populated buffer")
        kf = kfilter.kfilter(
            filter_length=3, filter_init=0, initial_guess=10, measurements=[10, 10, 10]
        )

        print("> should update OOR count when out of 0.5/2x range")
        kf.input_latest_measurement(20)
        self.assertEqual(kf.consecutive_replacements, 1)

        print("> should not truncate when (OOR count < filter_length)")
        kf.input_latest_measurement(21)
        self.assertEqual(kf.consecutive_replacements, 2)

        print("> should not truncate when (OOR count == filter_length)")
        kf.input_latest_measurement(22)
        self.assertEqual(kf.consecutive_replacements, 3)

        print(
            "> should truncate with too many consecutive replacement (OOR count > filter_length)"
        )
        kf.input_latest_measurement(25)
        self.assertEqual(len(kf.measurements), 2)
        self.assertEqual(kf.measurements[-1], 25)
        self.assertEqual(kf.consecutive_replacements, 0)

    def test_add_meas_part_buffer(self):
        print("\ntest_add_meas_part_buffer")
        print("initialize class with empty buffer")
        kf = kfilter.kfilter(
            filter_length=3, filter_init=0, initial_guess=10, measurements=[]
        )

        # print('{:06.3f}\t{:07.5f}\t{}\t{}'.format(kf.posteri_estimate, kf.posteri_error_estimate, kf.consecutive_replacements, kf.measurements))
        print("> test adding measurement, buffer 0/3")
        kf.input_latest_measurement(5)
        self.assertEqual(kf.measurements, [5])
        self.assertEqual(kf.get_latest_estimated_measurement(), (5, 5))

        print("> test adding measurement, buffer 1/3")
        kf.input_latest_measurement(5.1)
        self.assertEqual(kf.measurements, [5, 5.1])
        self.assertEqual(kf.get_latest_estimated_measurement(), (5.1, 5.1))

        print("> test adding measurement, buffer 2/3")
        kf.input_latest_measurement(5.2)
        self.assertEqual(kf.measurements, [5, 5.1, 5.2])
        self.assertEqual(kf.get_latest_estimated_measurement(), (5.2, 5.2))

        print("> test adding measurement, buffer 3/3")
        kf.input_latest_measurement(5.3)
        self.assertEqual(kf.measurements, [5.1, 5.2, 5.3])
        self.assertNotEqual(kf.get_latest_estimated_measurement(), (5.3, 5.3))

    def test_init_filter(self):
        print("\ntest_init_filter:")
        print("initialize filter with empty buffer")
        kf = kfilter.kfilter(filter_length=10, initial_guess=10, measurements=[])
        # disable init mode

        # print('{:06.3f}\t{:07.5f}\t{}\t{}'.format(kf.posteri_estimate, kf.posteri_error_estimate, kf.consecutive_replacements, kf.measurements))
        print("> filter is disabled during init period")
        for i in range(kf.filter_init):
            kf.input_latest_measurement(i)
            self.assertEqual(kf.measurements[-1], i)

        print("> filter should start after init period ends")
        kf.input_latest_measurement(i * 0.95)
        self.assertEqual(i * 0.95, kf.measurements[-1])
        self.assertNotEqual(kf.get_latest_estimated_measurement(), kf.measurements[-1])

    def test_blend(self):
        print("\ntest_run_blend:")
        # check blending is working as intended with default blending factor
        print("> filter should average signals 1 and 2 when noise = blend factor")
        self.assertEqual(kfilter.blend(1, 0, 0.05), 0.5)  # noise level = bf
        print("> filter should double-weight signal 2 when noise = 2 * blend factor")
        self.assertEqual(kfilter.blend(1, 0, 0.10), 1 / 3)  # noise level = 2x bf
        print("> filter should half-weight signal 2 when noise = 1/2 * blend factor")
        self.assertEqual(kfilter.blend(1, 0, 0.025), 2 / 3)  # noise level = 0.5x bf

        # adjust the blending factor
        print("> filter should adjust output based on the blend factor")
        self.assertEqual(kfilter.blend(1, 0, 0.05, blending_factor=0.10), 2 / 3)

    def test_run_psensefilter(self):
        print("\ntest_run_psensefilter:")

        # ts = (datetime.utcnow() - timedelta(minutes=10)).isoformat("T")
        data = pd.DataFrame(dict(signal1=[10] * 10, vout1=[0.4] * 10))

        print(
            "> filter executes kalman + mavg filters, and blends them based on SG accel"
        )
        res = kfilter.run_psensefilter(
            data, blending_factor=0.05, samples_mavg=100, samples_stdev=40
        )
        expected_keys = [
            "signal1",
            "vout1",
            "c_signal1",
            "fk_signal1",
            "f100_signal1",
            "d_signal1",
            "dd_signal1",
            "noise40_signal1",
            "fblend_signal1",
        ]
        self.assertTrue(res.keys().tolist() == expected_keys)

        res = kfilter.run_psensefilter(data, blending_factor=0.05)
        expected_keys = [
            "signal1",
            "vout1",
            "c_signal1",
            "fk_signal1",
            "f40_signal1",
            "d_signal1",
            "dd_signal1",
            "noise10_signal1",
            "fblend_signal1",
        ]
        self.assertTrue(res.keys().tolist() == expected_keys)

        self.assertEqual(res["signal1"][0], 10)
        self.assertEqual(res["signal1"][9], 10)
        self.assertEqual(res["vout1"][0], 0.4)
        self.assertEqual(res["c_signal1"][0], 10)
        self.assertEqual(res["d_signal1"][0], 0)
        self.assertEqual(res["d_signal1"].iloc[-1], 0)
        self.assertEqual(res["noise10_signal1"][0], 0)
        self.assertEqual(res["noise10_signal1"].iloc[-1], 0)
        self.assertEqual(res["fblend_signal1"][0], 10)
        self.assertEqual(res["fblend_signal1"].iloc[-1], 10)

        def frange(x, y, jump):
            while x <= y:
                yield x
                x += jump

        print("> filter response to a 0.01nA ramp")
        data = list(frange(5, 10, 0.01))
        data = pd.DataFrame(dict(signal1=data, vout1=[0.4] * len(data)))
        res = kfilter.run_psensefilter(
            data,
            blending_factor=1,
            samples_mavg=40,
            samples_stdev=7,
            kalman_coeff={
                "filter-window": 10,
                "filter-smoothing": 0.01,
                "filter-init": 40,
            },
        )
        self.assertEqual(res["signal1"].iloc[250], 7.5)
        self.assertEqual(res["fk_signal1"].iloc[250], 7.48)
        self.assertEqual(res["f40_signal1"].iloc[250], 7.31)
        self.assertEqual(res["d_signal1"].iloc[250], 0.01)
        self.assertEqual(res["fblend_signal1"].iloc[250], 7.49)
        self.assertEqual(res["fk_signal1"].iloc[-1], 9.98)
        self.assertEqual(res["f40_signal1"].iloc[-1], 9.81)
        self.assertEqual(res["d_signal1"].iloc[-1], 0.01)
        self.assertEqual(res["fblend_signal1"].iloc[-1], 9.99)

        print("> filter response to a 0.10nA ramp")
        data = list(frange(5, 10, 0.1))
        data = pd.DataFrame(dict(signal1=data, vout1=[0.4] * len(data)))
        res = kfilter.run_psensefilter(
            data,
            blending_factor=0.10,
            samples_mavg=40,
            samples_stdev=7,
            kalman_coeff={
                "filter-window": 10,
                "filter-smoothing": 0.05,
                "filter-init": 40,
            },
        )
        self.assertEqual(res["d_signal1"].iloc[-1], 0.09)
        self.assertEqual(res["fk_signal1"].iloc[-1], 9.7)
        self.assertEqual(res["fblend_signal1"].iloc[-1], 9.79)

        print("> filter response to a 0.10nA ramp, with blending factor modified")
        res = kfilter.run_psensefilter(
            data,
            blending_factor=0.10,
            samples_mavg=40,
            samples_stdev=7,
            kalman_coeff={
                "filter-window": 10,
                "filter-smoothing": 0.01,
                "filter-init": 40,
            },
        )
        self.assertEqual(res["fk_signal1"].iloc[-1], 9.35)
        self.assertEqual(res["fblend_signal1"].iloc[-1], 9.44)

        print("> filter response to a 1nA step change")

        def fstep(x, y, val1, val2):
            while x <= y:
                yield val1
                x += 1

            while x <= y + 30:
                yield val2
                x += 1

        data = list(fstep(1, 100, 5, 6))
        data = pd.DataFrame(dict(signal1=data, vout1=[0.4] * len(data)))
        res = kfilter.run_psensefilter(
            data,
            blending_factor=0.10,
            samples_mavg=40,
            samples_stdev=7,
            kalman_coeff={
                "filter-window": 10,
                "filter-smoothing": 0.01,
                "filter-init": 40,
            },
        )
        print("\t.. check kalman output")
        self.assertEqual(res["fk_signal1"].iloc[-31], 5)
        self.assertEqual(res["fk_signal1"].iloc[-30], 5.04)
        self.assertEqual(res["fk_signal1"].iloc[-21], 5.64)
        self.assertEqual(res["fk_signal1"].iloc[-11], 5.92)
        self.assertEqual(res["fk_signal1"].iloc[-1], 6)
        print("\t.. check blend output")
        self.assertEqual(res["fblend_signal1"].iloc[-31], 5)
        self.assertEqual(res["fblend_signal1"].iloc[-30], 5.05)
        self.assertEqual(res["fblend_signal1"].iloc[-10], 5.94)
        self.assertEqual(res["fblend_signal1"].iloc[-1], 6)


if __name__ == "__main__":
    unittest.main()



import numpy as np
import pandas as pd
import re
from datetime import datetime, timedelta
import json
import copy

# logging
import logging

log = logging.getLogger()


class StepResponse(object):
    def __init__(self, data, channels=[], c_thresh_ignore=0.25, debugmode=False):
        assert isinstance(data, pd.DataFrame), "data must be a pandas DataFrame"

        self.debug = debugmode
        self.msgid = "[StepResponse]"
        self.data = data
        self.to_per_minute = (
            60 / pd.Timedelta(np.median(np.diff(data.index))).total_seconds()
        )
        self.coeff_thresh_ignore_peak = c_thresh_ignore

        self.output = pd.DataFrame()
        self.result = dict()

        self.compute_sensor_response()

    def __str__(self):
        return json.dumps(self.result, indent=2)

    def is_valid(self):
        return len(self.output.index) > 0

    def compute_sensor_response(self, channels=[]):
        "collect response time related stats, return as dataframe (each row is a separate channel's results)"
        keys = self.data.columns.values.tolist()
        if len(channels) == 0:
            r = re.compile("signal*")
            channels = list(filter(r.match, keys))

        output = []
        for channel in channels:
            clip = self.data["c_{}".format(channel)]
            deriv = self.data["d_{}".format(channel)]

            if np.max(deriv) * self.to_per_minute < self.coeff_thresh_ignore_peak:
                log.debug(
                    "insufficient rate ({} na/sample). ignored".format(np.max(deriv))
                )
                continue

            clip_range = np.percentile(clip, [0, 100])
            cum_deriv = clip - clip[0]  # np.cumsum(deriv)
            goal_start = 0.1 * cum_deriv[-1]
            goal_end = 0.9 * cum_deriv[-1]

            t_rt10 = cum_deriv.where(cum_deriv < goal_start).last_valid_index()
            t_rt90 = cum_deriv.ge(goal_end).idxmax()
            t_rt = round((t_rt90 - t_rt10).total_seconds() / 6) / 10

            if t_rt < 0:
                log.debug("not a real step. ignoring.")
                continue

            start = self.data.index[0]
            finish = self.data.index[-1]

            self.result[channel] = dict(
                max_achieved_rate=np.max(deriv) * self.to_per_minute,
                response_time=t_rt,
                t10_ts=t_rt10.tz_localize(None).to_pydatetime().isoformat(),
                t10_val=clip[t_rt10].tolist(),
                t90_ts=t_rt90.tz_localize(None).to_pydatetime().isoformat(),
                t90_val=clip[t_rt90].tolist(),
            )

            output.append(
                [
                    start.tz_localize(None),
                    round((finish - start).total_seconds() / 6) / 10,
                    clip_range[0],
                    clip_range[1],
                    cum_deriv[-1],
                    t_rt,
                    t_rt10.tz_localize(None),
                    clip[t_rt10],
                    t_rt90.tz_localize(None),
                    clip[t_rt90],
                    np.max(deriv) * self.to_per_minute,
                ]
            )

            log.debug(
                "[channel={}] step analyzed, rt={:04.1f}m, dI={:05.2f}".format(
                    channel, t_rt, np.max(deriv) * self.to_per_minute
                )
            )

        # print(clip)
        self.output = pd.DataFrame.from_records(
            output,
            columns=[
                "T_Start",
                "Width_P",
                "I_P0",
                "I_P1",
                "Range_I_RT",
                "T_RT",
                "T_RT10",
                "I_RT10",
                "T_RT90",
                "I_RT90",
                "Max_dI",
            ],
        )


class PSenseResponse(object):
    "PercuSense class handles calculations associated with sensor response."

    def __init__(self, df, debugmode=False):
        self.debug = debugmode
        self.msgid = "[PSenseResponse]"
        self.data = df
        self.coeff_thresh_low = 0.012
        self.coeff_thresh_high = 2
        self.coeff_thresh_ignore_peak = 0.25
        self.result = dict()

    def find_steps_in_signal(self, signal, dsignal=None, numsteps=5, samplewindow=8):
        assert isinstance(signal, pd.Series), "signal should be a pandas Series"

        signal_min = (
            signal.quantile(0.01) if (len(signal) < 7000) else (signal.quantile(0.001))
        )
        signal_max = signal.quantile(0.95)
        thresh_too_high = (signal_max - signal_min) / (numsteps)
        thresh_high = thresh_too_high / self.coeff_thresh_high
        thresh_low = self.coeff_thresh_low * samplewindow

        log.info(
            "data content (channel {}): range({}, {})".format(
                signal.name, signal_max, signal_min
            )
        )
        log.debug("estimates based on {} steps".format(numsteps))
        log.debug(
            "thresholds: {:05.2f}, {:05.2f}, {:05.2f} nA".format(
                thresh_low, thresh_high, thresh_too_high
            )
        )

        # dsig = data['d_signal{}'.format(channel)]
        if dsignal is None:
            dsignal = signal.diff()

        rollsum = dsignal.rolling(samplewindow).sum()

        is_high = rollsum > thresh_high
        is_too_high = rollsum > thresh_too_high
        is_first_hit = is_high != (is_high.shift())  # or is_high != is_high.shift(2)

        steps = signal[is_high & ~is_too_high & is_first_hit]

        start_tm = []
        end_tm = []

        is_stable = signal[(rollsum > -thresh_low) & (rollsum < thresh_low)]
        for index, step in steps.iteritems():

            if len(end_tm) > 0 and is_stable[:index].index[-1] < end_tm[-1]:
                "don't allow overlapping steps"
                continue

            if (
                index >= is_stable.index[-1]
                or index < is_stable.index[0]
                or (is_stable[index:].index[0] - is_stable[:index].index[-1])
                < timedelta(seconds=15)
            ):
                "steps that are too rapid are likely noise"
                continue

            log.debug("new step identified {}: {}".format(index, step))
            start_tm.append(is_stable[:index].index[-1])
            end_tm.append(is_stable[index:].index[0])

        log.debug("found {} possible steps".format(len(start_tm)))

        return start_tm, end_tm

    def define_steps_in_signal(self, tm_start, time_window_mins=20):
        "manual definition of steps (alternative to find_steps_in_signal(..))"

        if isinstance(tm_start, list):
            tm_start = pd.Series(tm_start)

        assert isinstance(
            tm_start, pd.Series
        ), "tm_start must be of type list or pd.Series"

        # we want pandas Timestamps in our lists
        if not isinstance(tm_start[0], pd.Timestamp):
            tm_start = pd.to_datetime(tm_start)

        are_steps_independent = (
            tm_start.diff().min().total_seconds() / 60
        ) > time_window_mins
        assert (
            are_steps_independent
        ), "Overlapping steps detected. Timestamps must be at least time_window_mins ({} min) apart.".format(
            time_window_mins
        )

        # generate end-windows
        tm_finish = []
        for t0 in tm_start:
            tm_finish.append(t0 + timedelta(minutes=time_window_mins))

        return tm_start, pd.Series(tm_finish)

    def run(self, channels=[], tm_start=[], tm_finish=[], numsteps=5, samplewindow=8):
        "guess the steps and iterate through them to generate sensor response analyses for each"
        if len(channels) == 0:
            keys = self.data.columns.tolist()
            r = re.compile("signal*")
            channels = list(filter(r.match, keys))

        for channel in channels:
            log.info("analysis starting for channel {}".format(channel))
            if len(tm_start) > 0:
                assert (
                    tm_start[0].tzinfo is not None
                ), "timestamps should be timezone-aware"
            else:
                try:
                    tm_start, tm_finish = self.find_steps_in_signal(
                        signal=self.data["c_{}".format(channel)],
                        numsteps=numsteps,
                        samplewindow=samplewindow,
                    )
                except BaseException:
                    log.warning("find_steps_in_signal() failed. skipping sensor.")
                    continue

            self.result[channel] = pd.DataFrame()
            log.info("checking steps for sensor response")
            for start, finish in zip(tm_start, tm_finish):
                window = self.data[
                    (self.data.index >= start) & (self.data.index < finish)
                ]
                sr = StepResponse(
                    data=window,
                    c_thresh_ignore=self.coeff_thresh_ignore_peak,
                    channels=[channel],
                    debugmode=self.debug,
                )

                if sr.is_valid():
                    log.debug("adding step to result")
                    self.result[channel] = self.result[channel].append(
                        sr.output, ignore_index=True
                    )

    def getResult(self):
        return copy.deepcopy(self.result)



import unittest
from unittest.mock import patch, Mock
from decimal import Decimal
from datetime import datetime, timedelta
from psense_common import PSenseAWSInterface


class TestAWSItfc(unittest.TestCase):
    @patch("boto3.session.Session")
    def setUp(self, aws, test_mode=True):
        self.mock_table = Mock()
        self.mock_db = Mock()
        self.mock_db.Table.return_value = self.mock_table
        # self.mock_db.batch_write_item.return_value = dict()
        aws().resource.return_value = self.mock_db

        # generate a default response to Table methods
        self.mock_table.query.return_value = dict(
            Count=1, Items=[{"timestamp": datetime.utcnow().isoformat()}]
        )
        self.mock_table.put_item.return_value = dict()
        self.mock_table.delete_item.return_value = "{}"

        # generate a mock batch_writer(), attach it to mock_table.batch_writer()
        self.mock_batch_writer = Mock()
        self.mock_batch_writer.__enter__ = Mock(return_value=self.mock_batch_writer)
        self.mock_batch_writer.__exit__ = Mock(return_value=None)
        self.mock_table.batch_writer.return_value = self.mock_batch_writer

        self.aws = PSenseAWSInterface(testmode=test_mode, debugmode=True)
        self.test_expid = "19A01A001F01-0001-001-00-G01CPVX"

    def test_class_init(self):
        obj = PSenseAWSInterface(profile_name="doesnotexist")
        self.assertEqual(obj.aws.profile_name, "default")
        self.assertEqual(obj.aws.region_name, "us-east-1")
        self.assertEqual(obj.experiment_name, "")
        self.assertEqual(obj.experiment_info, [])
        self.assertFalse(obj.exp_is_valid)

    def test_set_query_config(self):
        self.aws.set_query_config(1, 2)
        self.assertEqual(self.aws.query_req_limit, 1)
        self.assertEqual(self.aws.query_req_count_limit, 2)

    def test_setget_expid(self):
        self.aws.set_experiment("asdf")
        self.assertEqual(self.aws.experiment_name, "asdf")
        self.assertEqual(self.aws.get_experiment(), "asdf")
        self.aws.set_experiment("")

    def test_verif_experiment(self):
        # experiment not verified if it is None
        self.assertFalse(self.aws.verif_experiment())

        # allow extended-length device ids
        self.aws.verif_experiment("19A01A001M9999-987A-654-3Z-213CCVACW")
        self.assertTrue(
            self.aws.verif_experiment("19A01A001M9999-987A-654-3Z-213CCVACW")
        )
        self.assertTrue(self.aws.exp_is_valid)

        # check parsing is done properly on verification
        self.assertTrue(self.aws.verif_experiment(self.test_expid))
        self.assertTrue(self.aws.exp_is_valid)
        self.assertEqual(self.aws.experiment_info["study"]["start"], "2019-01-01")
        self.assertEqual(self.aws.experiment_info["sensor"]["vsn"], "01")

        self.assertTrue(self.aws.verif_experiment("20H07R301M99-0400-400-00-G15CPVX"))
        self.assertTrue(self.aws.verif_experiment("20H07R301M99-0400-400-00-299OCWGRV"))

        self.assertTrue(self.aws.verif_experiment("20H07R301M99-0400-400-00-111GCV"))
        self.assertFalse(
            self.aws.verif_experiment("20H07R301M99-0400-400-00-115GCVGCV")
        )
        self.assertTrue(
            self.aws.verif_experiment("20H07R301M99-0400-400-00-315GCVGCVGCV")
        )
        self.assertFalse(
            self.aws.verif_experiment("20H07R301M99-0400-400-00-315GCVGCV")
        )

        self.assertTrue(self.aws.verif_experiment("20H07R301M99-9999-ABC-DE-215ACSCCB"))
        self.assertFalse(self.aws.verif_experiment("20H07R301M9-0400-400-00-215GCVGCV"))
        self.assertFalse(
            self.aws.verif_experiment("20H07R301M99-04000400-00-215GCVGCV")
        )
        self.assertFalse(self.aws.verif_experiment("20H07R301M99-0400-00-00-215GCVGCV"))

    # @patch('boto3.session.Session.resource')
    def test_check_experiment_data(self):
        self.setUp()  # reinitialize

        self.aws.test_mode = True
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(valid, "Yes")
        self.assertTrue(isactive)
        self.assertEqual(vout, -999)

        self.aws.test_mode = False
        self.mock_table.query.return_value = dict(
            Count=1, Items=[{"timestamp": datetime.utcnow().isoformat(), "vout1": 1}]
        )
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(valid, "Yes")
        self.assertTrue(isactive)
        self.assertEqual(vout, "1.000")

        self.mock_table.query.return_value = dict(
            Count=0,
            Items=[
                {
                    "timestamp": (datetime.utcnow() - timedelta(days=1)).isoformat(),
                    "vout1": 2,
                }
            ],
        )
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(valid, "No")
        self.assertFalse(isactive)
        self.assertEqual(vout, "0")

        self.mock_table.query.return_value = dict(
            Count=1,
            Items=[
                {
                    "timestamp": (datetime.utcnow() - timedelta(days=1)).isoformat(),
                    "vout1": 1,
                    "vout2": 2,
                    "vout3": 3,
                    "enable1": True,
                    "enable2": False,
                    "enable3": True,
                }
            ],
        )
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(vout, "1.000 / (off) / 3.000")

        self.aws.test_mode = True

    # @patch('boto3.session.Session.resource')
    def test_multiple_vouts_in_data(self):
        self.setUp()  # reinitialize

        self.aws.test_mode = False
        self.mock_table.query.return_value = dict(
            Count=1,
            Items=[
                {
                    "timestamp": (datetime.utcnow() - timedelta(days=1)).isoformat(),
                    "vout1": 0.5,
                    "vout2": 0.4999,
                    "vout3": 3,
                }
            ],
        )
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(valid, "Yes")
        self.assertFalse(isactive)
        self.assertEqual(vout, "0.500 / 0.500 / 3.000")

        # check that vout keys are sorted correctly
        self.mock_table.query.return_value = dict(
            Count=1,
            Items=[
                {
                    "timestamp": (datetime.utcnow() - timedelta(days=1)).isoformat(),
                    "vout3": 3,
                    "vout2": 2,
                    "vout1": 1,
                }
            ],
        )
        valid, isactive, vout = self.aws.check_experiment_data("test")
        self.assertEqual(valid, "Yes")
        self.assertFalse(isactive)
        self.assertEqual(vout, "1.000 / 2.000 / 3.000")
        self.aws.test_mode = True

    def test_aws_query_timeout(self):
        from botocore.exceptions import ReadTimeoutError

        self.setUp()
        self.aws.test_mode = False
        self.aws.verif_experiment(self.test_expid)

        # generate fake timeouts in response to a query
        self.mock_table.query.side_effect = ReadTimeoutError(
            endpoint_url="https://localhost"
        )

        # add_experiment() handled properly
        response, msg = self.aws.add_experiment()
        self.assertEqual(msg, "timeout")
        self.assertFalse(response)

        # check_experiment_data
        self.assertEqual(
            self.aws.check_experiment_data(self.test_expid), (None, None, None)
        )

        # get_event_data
        self.assertEqual(self.aws.get_event_data(), None)

        # get_sensordata_slice
        self.assertEqual(self.aws.get_sensordata_slice("2000", 1), (None, None))

        # get_sensordata_sincetimestamp
        self.assertEqual(self.aws.get_sensordata_sincetimestamp("2000"), (None, None))

        # get_sensor_data
        self.assertEqual(self.aws.get_sensor_data(), (None, None))

        # get_experiments
        self.assertEqual(self.aws.get_experiments(), [])

    def test_add_experiment(self):
        self.setUp()  # reinitialize
        # fail if no expid
        self.assertRaises(AssertionError, self.aws.add_experiment)
        self.aws.set_experiment(self.test_expid)
        # fail if expid has not been verified
        self.assertRaises(AssertionError, self.aws.add_experiment)
        self.aws.verif_experiment()

        # if in testmode, return true
        self.aws.test_mode = True
        res, msg = self.aws.add_experiment()
        self.assertTrue(res)

        # not in test mode, we need to mock the response
        self.aws.test_mode = False

        # handle aws success, but experiment already exists
        self.mock_table.query.return_value = dict(
            Count=1, Items=[{"timestamp": datetime.utcnow().isoformat()}]
        )
        res, msg = self.aws.add_experiment()
        self.assertFalse(res)
        self.assertEqual(msg, "exists")

        # handle aws success
        self.mock_table.query.return_value = dict(Count=0, Items=[])
        res, msg = self.aws.add_experiment()
        self.assertTrue(res)
        self.assertEqual(msg, "")

        # handle aws failed
        self.mock_table.put_item.side_effect = Exception("Test")
        res, msg = self.aws.add_experiment()
        self.assertFalse(res)
        self.assertEqual(msg, "error")

        self.aws.test_mode = True

    def test_add_sensordata(self):
        self.setUp()
        self.assertRaises(AssertionError, self.aws.add_sensordata)

        self.aws.verif_experiment(self.test_expid)

        self.aws.test_mode = True
        self.assertTrue(self.aws.add_sensordata(vout=[0], signal=[0]))
        self.assertTrue(self.aws.add_sensordata(signal=[1, 2]))
        self.assertTrue(
            self.aws.add_sensordata(timestamp=datetime.utcnow(), vout=[0], signal=[0])
        )
        self.assertTrue(
            self.aws.add_sensordata(timestamp=datetime.now(), vout=[0], signal=[0])
        )
        self.assertTrue(
            self.aws.add_sensordata(
                timestamp=datetime.now().replace(tzinfo=None), vout=[0], signal=[0]
            )
        )
        self.assertTrue(
            self.aws.add_sensordata(
                timestamp=datetime.now().isoformat("T"), vout=[0], signal=[0]
            )
        )

        self.aws.test_mode = False
        self.assertTrue(self.aws.add_sensordata(vout=[0], signal=[0]))
        self.mock_table.put_item.side_effect = Exception("Test")
        self.assertFalse(self.aws.add_sensordata(vout=[0], signal=[0]))
        self.aws.test_mode = True

    def test_generate_put_request(self):
        self.setUp()
        self.assertRaises(AssertionError, self.aws.generate_put_request, [])

        self.aws.verif_experiment(self.test_expid)
        check_timestamp = datetime.utcnow().isoformat()
        request = self.aws.generate_put_request([check_timestamp, 0.5, 0, True])
        self.assertEqual(request["experiment"], self.test_expid)
        self.assertEqual(request["vout1"], Decimal(0.5))
        self.assertEqual(request["signal1"], Decimal(0))
        self.assertTrue(request["enable1"])

        # try multi-channel data
        test_data = [check_timestamp, 0.5, 0, False, 1.5, 1, True]
        request = self.aws.generate_put_request(test_data)
        self.assertEqual(request["experiment"], self.test_expid)
        self.assertEqual(request["timestamp"], check_timestamp)
        self.assertEqual(request["vout1"], Decimal(0.5))
        self.assertEqual(request["signal1"], Decimal(0))
        self.assertFalse(request["enable1"])
        self.assertEqual(request["vout2"], Decimal(1.5))
        self.assertEqual(request["signal2"], Decimal(1))
        self.assertTrue(request["enable2"])

    @patch("time.sleep", return_value=None)
    def test_add_bulk_sensordata(self, delay=None):
        self.setUp()
        self.assertRaises(AssertionError, self.aws.add_bulk_sensordata, [])

        self.aws.verif_experiment(self.test_expid)

        data = [
            [datetime.utcnow().isoformat(), 1, 0.1, True],
            [datetime.utcnow(), 2, 0.2, True],
            [datetime.now(), 3, 0.3, True],
        ]

        self.aws.test_mode = True
        self.assertTrue(self.aws.add_bulk_sensordata(data))
        self.aws.batch_request_limit = 1

        extradata = [[datetime.now(), 3, 0.3, True] for i in range(26)]
        self.assertTrue(self.aws.add_bulk_sensordata(extradata))

        self.aws.test_mode = False
        self.mock_batch_writer._response = dict(
            ResponseMetadata=dict(HTTPStatusCode=200)
        )
        self.assertTrue(self.aws.add_bulk_sensordata(data))
        self.assertTrue(self.aws.add_bulk_sensordata(extradata))

        self.mock_batch_writer._response = dict(
            ResponseMetadata=dict(HTTPStatusCode=301, RetryAttempts=1),
            UnprocessedItems=dict(),
        )
        self.assertFalse(self.aws.add_bulk_sensordata(extradata))
        self.mock_batch_writer.put_item.side_effect = Exception("Test")

    def test_get_experiments(self):
        self.setUp()
        self.mock_table.query.return_value = dict(
            Count=2,
            Items=[{"experiment": "Experiment1"}, {"experiment": "Experiment2"}],
        )
        ret = self.aws.get_experiments()
        self.assertTrue(isinstance(ret, list))
        self.assertEqual(ret[0], "Experiment1")
        self.assertEqual(ret[1], "Experiment2")
        self.mock_table.query.return_value = dict(
            Count=2,
            Items=[{"experiment": "Experiment2"}, {"experiment": "Experiment1"}],
        )
        self.assertEqual(ret[0], "Experiment1")
        self.assertEqual(ret[1], "Experiment2")

    def test_get_sensor_data(self):
        self.setUp()
        self.aws.verif_experiment(self.test_expid)
        check_timestamp = datetime.utcnow()
        count, request = self.aws.get_sensor_data()
        self.assertEqual(count, 1)
        self.assertEqual(request["signal1"][0], 10)
        self.assertEqual(request["vout1"][0], 0.4)

        self.aws.test_mode = False
        self.mock_table.query.return_value = {
            "Count": 1,
            "Items": [
                {
                    "experiment": self.test_expid,
                    "timestamp": check_timestamp.isoformat("T"),
                    "signal1": 20,
                    "vout1": 0.4,
                    "enable1": True,
                }
            ],
        }
        count, request = self.aws.get_sensor_data()

        self.assertEqual(count, 1)
        self.assertEqual(request["signal1"][0], 20)
        self.assertEqual(request["vout1"][0], 0.4)
        self.assertTrue(request["enable1"][0])

        self.aws.set_query_config(req_size=2, req_count=1)
        self.mock_table.query.return_value = {
            "Count": 2,
            "Items": [
                {
                    "experiment": self.test_expid,
                    "timestamp": (check_timestamp - timedelta(minutes=10)).isoformat(
                        "T"
                    ),
                    "signal1": 20,
                    "vout1": 0.4,
                    "enable1": True,
                }
            ]
            * 2,
        }
        count, request = self.aws.get_sensor_data()
        self.assertEqual(count, 2)
        self.assertEqual(request["signal1"][0], 20)
        self.assertEqual(request["vout1"][0], 0.4)
        self.assertTrue(request["enable1"][0])

        self.aws.set_query_config(req_size=2, req_count=2)
        count, request = self.aws.get_sensor_data()
        self.assertEqual(count, 4)
        self.assertEqual(request["signal1"][0], 20)
        self.assertEqual(request["vout1"][0], 0.4)
        self.assertTrue(request["enable1"][0])

        self.aws.set_query_config(req_size=2, req_count=3)
        self.mock_table.query.side_effect = [
            {
                "Count": 2,
                "Items": [
                    {
                        "experiment": self.test_expid,
                        "timestamp": (
                            check_timestamp - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 10,
                        "vout1": 0.4,
                        "enable1": True,
                    }
                ]
                * 2,
            },
            {
                "Count": 2,
                "Items": [
                    {
                        "experiment": self.test_expid,
                        "timestamp": (
                            check_timestamp - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 10,
                        "vout1": 0.4,
                        "enable1": True,
                    }
                ]
                * 2,
            },
            {
                "Count": 1,
                "Items": [
                    {
                        "experiment": self.test_expid,
                        "timestamp": (
                            check_timestamp - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 5,
                        "vout1": 0.1,
                        "enable1": False,
                    }
                ],
            },
        ]
        count, request = self.aws.get_sensor_data()
        self.assertEqual(count, 5)
        self.assertEqual(request["signal1"][0], 10)
        self.assertEqual(request["vout1"][0], 0.4)
        self.assertTrue(request["enable1"][0])

        self.assertEqual(request["signal1"][-1], 5)
        self.assertEqual(request["vout1"][-1], 0.1)
        self.assertFalse(request["enable1"][-1])

        self.aws.test_mode = True

    def test_get_sensor_data_multianalyte(self):
        self.setUp()
        self.aws.verif_experiment(self.test_expid)
        self.aws.test_mode = False
        self.mock_table.query.return_value = {
            "Count": 1,
            "Items": [
                {
                    "experiment": self.test_expid,
                    "timestamp": (datetime.utcnow() - timedelta(minutes=10)).isoformat(
                        "T"
                    ),
                    "signal1": 10,
                    "vout1": 0.4,
                    "enable1": True,
                    "signal2": 20,
                    "vout2": 0.6,
                    "enable2": True,
                    "signal3": -30,
                    "vout3": -0.5,
                    "enable3": False,
                }
            ],
        }
        count, request = self.aws.get_sensor_data()
        self.aws.test_mode = True

        self.assertEqual(count, 1)
        self.assertEqual(request["signal1"][0], 10)
        self.assertEqual(request["signal2"][0], 20)
        self.assertEqual(request["signal3"][0], -30)

        self.assertEqual(request["vout1"][0], 0.4)
        self.assertEqual(request["vout2"][0], 0.6)
        self.assertEqual(request["vout3"][0], -0.5)

        self.assertTrue(request["enable1"][0])
        self.assertTrue(request["enable2"][0])
        self.assertFalse(request["enable3"][0])

    def test_get_sensordata_slice(self):
        self.setUp()
        self.aws.verif_experiment(self.test_expid)
        check_timestamp = datetime.utcnow()

        self.aws.test_mode = True
        count, request = self.aws.get_sensordata_slice(
            check_timestamp.isoformat("T"), 5
        )
        self.assertEqual(count, 1)
        self.assertEqual(request["signal1"][0], 5)
        self.assertEqual(request["vout1"][0], 0.2)
        self.assertTrue(request["enable1"][0])

        self.aws.test_mode = False
        self.mock_table.query.return_value = {
            "Count": 1,
            "Items": [
                {
                    "experiment": self.test_expid,
                    "timestamp": (check_timestamp - timedelta(minutes=10)).isoformat(
                        "T"
                    ),
                    "signal1": 6,
                    "vout1": 0.3,
                    "enable1": True,
                }
            ],
        }
        count, request = self.aws.get_sensordata_slice(check_timestamp, 5)
        self.assertEqual(count, 1)
        self.assertEqual(request["signal1"][0], 6)
        self.assertEqual(request["vout1"][0], 0.3)

        self.mock_table.query.return_value = {
            "Count": 3,
            "Items": [
                {
                    "experiment": self.test_expid,
                    "timestamp": (check_timestamp - timedelta(minutes=10)).isoformat(
                        "T"
                    ),
                    "signal1": 10,
                    "vout1": 0.4,
                    "enable1": True,
                }
            ]
            * 3,
        }
        count, request = self.aws.get_sensordata_slice(check_timestamp, 5)
        self.assertEqual(count, 3)

        self.mock_table.query.return_value = {"Count": 0, "Items": []}
        count, request = self.aws.get_sensordata_slice(check_timestamp, 5)
        self.assertEqual(count, 0)

        self.aws.test_mode = True

    def test_get_event_data(self):
        self.setUp()
        self.aws.set_experiment(self.test_expid)
        self.assertRaises(AssertionError, self.aws.get_event_data)
        self.aws.verif_experiment()

        self.aws.test_mode = True
        self.assertEqual(len(self.aws.get_event_data()), 1)

    def test_add_event(self):
        self.setUp()
        self.aws.set_experiment(self.test_expid)
        self.assertRaises(AssertionError, self.aws.add_event, 0, "", "", 0, 0)
        self.aws.verif_experiment()
        self.assertRaises(
            AssertionError, self.aws.add_event, 0, "", "not-a-real-type", 0, 0
        )

        self.aws.test_mode = True
        self.assertTrue(self.aws.add_event(0, datetime.now(), "Lactate", 0, 0))
        self.assertTrue(
            self.aws.add_event(0, datetime.utcnow().isoformat(), "Lactate", 0, 0)
        )

    def test_delete_event(self):
        self.setUp()
        self.aws.set_experiment(self.test_expid)
        self.assertRaises(AssertionError, self.aws.delete_event, 0, "", "")
        self.aws.verif_experiment()
        self.assertRaises(
            AssertionError, self.aws.delete_event, 0, "", "not-a-real-type"
        )
        self.aws.test_mode = True
        self.assertTrue(self.aws.delete_event(0, datetime.now(), "Lactate"))
        self.assertTrue(
            self.aws.delete_event(0, datetime.utcnow().isoformat(), "Lactate")
        )
        self.aws.test_mode = False
        self.assertEqual(self.aws.delete_event(0, datetime.now(), "Lactate"), dict())
        self.aws.test_mode = True


if __name__ == "__main__":
    unittest.main()



import pandas as pd
from psense_common import PSenseParser
from psense_common.psense_parser import is_float, SensorDataFormat
import unittest
from unittest.mock import patch, MagicMock, Mock
from datetime import datetime, timedelta
import pytz


def is_dst():
    """Determine whether or not Daylight Savings Time (DST)
    is currently in effect"""

    x = datetime(
        datetime.now().year, 1, 1, 0, 0, 0, tzinfo=pytz.timezone("US/Eastern")
    )  # Jan 1 of this year
    y = datetime.now(pytz.timezone("US/Eastern"))

    # if DST is in effect, their offsets will be different
    return not (y.utcoffset() == x.utcoffset())


class TestPSenseParser(unittest.TestCase):
    def setUp(self):
        self.parser = PSenseParser(debugmode=True)
        pass

    def test_is_float(self):
        self.assertTrue(is_float(0))
        self.assertTrue(is_float("0"))
        self.assertTrue(is_float("0.5"))
        self.assertFalse(is_float("A"))
        self.assertFalse(is_float({"value": 0.5}))
        self.assertFalse(is_float([0.5]))

    def test_sensordataformat(self):
        obj = SensorDataFormat()

        # empty init
        self.assertIsNone(obj.name)
        self.assertEqual(obj.delimiter, ",")
        self.assertIsNone(obj.type_is_sensor)
        self.assertEqual(obj.has_timestamp, True)

        self.assertIsNone(obj.get_key("signal1"))
        self.assertIsNone(obj.get_header("signal1"))
        self.assertIsNone(obj.get_index("signal1"))
        self.assertEqual(obj.data_headers(), [])
        self.assertEqual(obj.data_keys(), [])
        self.assertEqual(obj.data_indices(), [])

        obj.set_key("signal1", ("test", 20))
        self.assertEqual(obj.get_key("signal1"), ("test", 20))
        self.assertEqual(obj.get_header("signal1"), "test")
        self.assertEqual(obj.get_index("signal1"), 20)

        self.assertEqual(obj.data_headers(), ["test"])
        self.assertEqual(obj.data_keys(), ["signal1"])
        self.assertEqual(obj.data_indices(), [20])

        obj = SensorDataFormat(
            name="test",
            delimiter="\t",
            has_timestamp=False,
            columns=dict(
                type=("TYPE", 1),
                timestamp=("timestamp", 0),
                signal1=("SIGNAL1", 2),
                vout1=("VOUT1", 4),
                enable1=("enable1", 3),
            ),
        )
        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.delimiter, "\t")
        self.assertEqual(obj.has_timestamp, False)
        self.assertEqual(obj.get_key("signal1"), ("SIGNAL1", 2))
        self.assertIsNone(obj.get_key("doesntexist"))
        self.assertEqual(obj.get_header("signal1"), "SIGNAL1")
        self.assertEqual(obj.get_index("signal1"), 2)
        self.assertListEqual(
            obj.data_headers(), ["timestamp", "TYPE", "SIGNAL1", "enable1", "VOUT1"]
        )
        self.assertListEqual(
            obj.data_keys(), ["timestamp", "type", "signal1", "enable1", "vout1"]
        )
        self.assertListEqual(obj.data_indices(), [1, 0, 3, 4, 2])

    def test_force_source(self):
        self.assertRaises(IOError, self.parser.force_source, "not-a-real-data-type")
        type1 = "PSHIELD"
        self.parser.force_source(type1)
        self.assertEqual(self.parser.source, type1)
        type2 = "BWII-DL"
        self.parser.force_source(type2)
        self.assertEqual(self.parser.source, type2)

    def test_find_header_text(self):
        mock_type = "PSHIELD"
        search_string = self.parser.header_text[mock_type]

        m_open = MagicMock()
        m_methods = Mock()
        m_methods.tell.side_effect = [1, 2, 3, 4, 5, 15, 15]
        m_methods.readline.side_effect = [
            "random",
            "random",
            search_string,
            "random",
            "random",
            "random",
        ]
        m_open().__enter__.return_value = m_methods

        with patch("builtins.open", m_open, create=True):
            with patch("os.path.getsize", return_value=11):
                self.assertTrue(
                    self.parser.find_header_text("fake_file", search_string)
                )
                self.assertFalse(
                    self.parser.find_header_text(
                        "fake_file", "string-that-doesnt-exist"
                    )
                )

    def test_identify_file_source(self):

        self.assertRaises(
            FileNotFoundError, self.parser.identify_file_source, "fake_file"
        )

        self.parser.identify_file_source(fpath="tests/pshield_example")
        self.assertEqual(self.parser.source, "PSHIELD")
        self.parser.identify_file_source(fpath="tests/gamry_example")
        self.assertEqual(self.parser.source, "EXPLAIN")
        self.parser.identify_file_source("tests/__init__.py")
        self.assertEqual(self.parser.source, None)

    def test_read_variable_header_file(self):
        data, header = self.parser.read_variable_header_file(
            "tests/__init__.py", "NOT_FOUND_HEADER", sep=","
        )
        self.assertIsNone(data)
        self.assertIsNone(header)

        data, header = self.parser.read_variable_header_file(
            "tests/web1chan_example", "timestamp,Raw1,Vout1", sep=","
        )
        self.assertIsInstance(data, pd.DataFrame)
        self.assertListEqual(
            list(data.columns), ["timestamp", "Raw1", "Vout1", "Filt1"]
        )
        self.assertEqual(data.shape, (10, 4))
        self.assertEqual(len(header), 190)

    def test_parse_record(self):
        starting_time = datetime.now()

        self.parser.force_source("BWII-MINI")
        data = ",".join(
            [
                starting_time.isoformat(),
                "sensor_record",
                "10",
                "1.0",
                "1000",
                "500",
                "1",
                "0",
                "2.0",
                "1010",
                "1",
                "0",
                "3.0",
                "1500",
                "1",
                "0",
                "3300",
            ]
        )
        res = self.parser.parse_record(data)
        self.assertTrue(
            datetime.fromisoformat(res[0]) - starting_time < timedelta(seconds=3)
        )
        self.assertEqual(res[1], [0.5, 0.51, 1.0])
        self.assertEqual(res[2], [1.0, 2.0, 3.0])
        self.assertEqual(res[3], [True, True, True])
        data = ",".join(
            [
                starting_time.isoformat(),
                "sensor_record",
                "10",
                "1.0",
                "1000",
                "500",
                "0",
                "0",
                "2.0",
                "1010",
                "0",
                "0",
                "3.0",
                "1500",
                "0",
                "0",
                "3300",
            ]
        )
        res = self.parser.parse_record(data)
        self.assertEqual(res[3], [False, False, False])

        self.parser.force_source("BWII-DL")
        data = ",".join(
            [
                starting_time.isoformat(),  # timestamp
                "sensor_record",  # type
                "1",  # rnum
                "1",  # i
                ".2",  # vw
                ".1",  # vr
                "0",  # vc
                "1",  # on
                "0",  # disc
                "10",  # i
                ".4",  # vw
                ".2",  # vr
                ".1",  # vc
                "1",  # on
                "0",  # disc
                "0",  # batt
                "3260",  # vbatt
            ]
        )
        res = self.parser.parse_record(data)
        self.assertTrue(
            datetime.fromisoformat(res[0]) - starting_time < timedelta(seconds=3)
        )
        self.assertEqual(res[1], [0.0001, 0.0002])
        self.assertEqual(res[2], [1.0, 10.0])
        self.assertEqual(res[3], [True, True])
        data = ",".join(
            [
                starting_time.isoformat(),
                "sensor_record",  # type
                "1",  # rnum
                "1",  # i1
                ".2",  # vw
                ".1",  # vr
                "0",  # vc
                "0",  # on
                "0",  # disc
                "10",  # i
                ".4",  # w
                ".2",  # r
                ".1",  # c
                "0",  # on
                "0",  # disc
                "0",  # batt
                "3260",  # vbatt
            ]
        )
        res = self.parser.parse_record(data)
        self.assertEqual(res[3], [False, False])

        data = "1, 1554417874.0000026, 12.51"
        self.parser.force_source("PSHIELD")
        res = self.parser.parse_record(data)
        self.assertEqual(res[0], "2019-04-04T14:44:34.000003")  # local time, not utc
        self.assertEqual(res[1], [None])
        self.assertEqual(res[2], [12.51])
        self.assertEqual(res[3], [True])

        data = "0.399550	7.3E-9"
        self.parser.force_source("VFP")
        res = self.parser.parse_record(data)
        self.assertTrue(
            datetime.fromisoformat(res[0]) - starting_time < timedelta(seconds=3)
        )
        self.assertEqual(res[1], [0.39955])
        self.assertEqual(res[2], [7.3])
        self.assertEqual(res[3], [True])

        data = "2015-12-26 20:19:18,0.72,0.3,0.72"
        self.parser.force_source("WEB-APP")
        res = self.parser.parse_record(data)
        self.assertEqual(res[0], "2015-12-26T20:19:18")
        self.assertEqual(res[1], [0.3])
        self.assertEqual(res[2], [0.72])
        self.assertTrue(res[3])

        data = "2020-02-10 10:59:24,0.83,0.507,-7.89,-0.456,5.05,0.123,0.83,1.3,0.03"
        self.parser.force_source("WEB-APP")
        res = self.parser.parse_record(data)
        self.assertEqual(res[0], "2020-02-10T10:59:24")
        self.assertEqual(res[1], [0.507, -0.456, 0.123])
        self.assertEqual(res[2], [0.83, -7.89, 5.05])
        self.assertEqual(res[3], [True] * 3)

        res = self.parser.parse_record(data, num_channels=2)
        self.assertEqual(res[0], "2020-02-10T10:59:24")
        self.assertEqual(res[1], [0.507, -0.456])
        self.assertEqual(res[2], [0.83, -7.89])
        self.assertEqual(res[3], [True] * 2)

        data = "1.000e+1, 4.020e-8, 5.779e-8, 6.402e-8, 7.063e-8"
        self.parser.force_source("CHI")
        res = self.parser.parse_record(data)
        self.assertEqual(res[1], [None] * 4)
        self.assertEqual(res[2], [4.02e-8, 5.779e-8, 6.402e-8, 7.063e-8])
        self.assertEqual(res[3], [True] * 4)
        data = "1.000e+1, 1e-8, 2e-8, 3e-8, 4e-8, -1e-8, -2e-8, -3e-8, -4e-8"
        res = self.parser.parse_record(data)
        self.assertEqual(res[1], [None] * 8)
        self.assertEqual(res[2], [1e-8, 2e-8, 3e-8, 4e-8, -1e-8, -2e-8, -3e-8, -4e-8])
        self.assertEqual(res[3], [True] * 8)

    def test_load_no_source(self):
        testfile = "tests/pshield_example"

        # fake source
        self.parser.source = "fake-source"
        self.assertRaises(ValueError, self.parser.load_rawfile, testfile, None, None)

    def test_load_pshield(self):
        testfile = "tests/pshield_example"
        # pshield file
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[-1], 0.5)
        self.assertEqual(df["signal1"].iloc[-1], 5.29)
        self.assertEqual(
            df["timestamp"].iloc[-1], pd.Timestamp("2018-07-23 16:28:08.229594946")
        )

    def test_load_bwii_script(self):
        # psense format (bwii-instrument.py)
        testfile = "tests/psense_example"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[-1], 0)
        self.assertEqual(df["signal1"].iloc[-1], 1)
        self.assertEqual(df["vout2"].iloc[-1], 2)
        self.assertEqual(df["signal2"].iloc[-1], 3)
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-03-04T13:08:54"))

        self.assertTrue(df["enable1"].iloc[-1])
        self.assertTrue(df["enable2"].iloc[0])

    def test_load_bwii_dl(self):
        # bwii format (bwii-dl)
        testfile = "tests/bwii_example"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[-1], 0.487)
        self.assertEqual(df["signal1"].iloc[-1], 1.76)
        self.assertEqual(df["vout2"].iloc[-1], 0)
        self.assertEqual(df["signal2"].iloc[-1], 0)
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-03-04 16:37:42"))

        start_time = "2019/01/01 12:00"
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, False))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[-1], 0.487)
        self.assertEqual(df["signal1"].iloc[-1], 1.76)
        self.assertEqual(df["vout2"].iloc[-1], 0)
        self.assertEqual(df["signal2"].iloc[-1], 0)
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-01-01 12:13:43"))

        self.assertTrue(self.parser.load_rawfile(testfile, start_time, True))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[-1], 0.487)
        self.assertEqual(df["signal1"].iloc[-1], 1.76)
        self.assertEqual(df["vout2"].iloc[-1], 0)
        self.assertEqual(df["signal2"].iloc[-1], 0)
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-01-01 12:00:00"))

    def test_load_bwii_mini(self):
        testfile = "tests/bwii_mini_example"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(df["vout1"].iloc[1], 0.501)
        self.assertEqual(df["signal1"].iloc[1], 0.175)
        self.assertEqual(df["vout2"].iloc[1], 0.502)
        self.assertEqual(df["signal2"].iloc[1], 1.75)
        self.assertEqual(df["vout3"].iloc[1], 1)
        self.assertEqual(df["signal3"].iloc[1], 17.5)

        self.assertEqual(df["vout1"].iloc[-1], 0.501)
        self.assertEqual(df["signal1"].iloc[-1], 175)
        self.assertEqual(df["vout2"].iloc[-1], -0.488)
        self.assertEqual(df["signal2"].iloc[-1], 0)
        self.assertEqual(df["vout3"].iloc[-1], -0.488)
        self.assertEqual(df["signal3"].iloc[-1], 0)
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2020-01-01 19:32:17"))

        self.assertTrue(df["enable1"].iloc[0])
        self.assertFalse(df["enable2"].iloc[0])
        self.assertFalse(df["enable3"].iloc[0])
        self.assertFalse(df["enable3"].iloc[-1])

        start_time = "2019/01/01 12:00"
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, False))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp("2019-01-01 12:00:00"))
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-01-01 12:02:19"))

        self.assertTrue(self.parser.load_rawfile(testfile, start_time, True))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp("2019-01-01 11:57:41"))
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-01-01 12:00:00"))

    def test_load_bwii_util(self):
        # files downloaded from BWII_Util.exe have extra whitespace. Make sure they are loaded properly
        testfile = "tests/bwii_util_example"
        self.parser.identify_file_source(testfile)
        self.assertEqual(self.parser.source, "BWII-MINI")
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))

        df = self.parser.data
        self.assertEqual(len(df.index), 20)
        self.assertEqual(df["signal1"].iloc[0], 41.5901)
        self.assertEqual(df["signal2"].iloc[0], 0.0520)
        self.assertEqual(df["signal3"].iloc[0], 0.0500)
        self.assertEqual(df["signal1"].iloc[-1], 9.5240)
        self.assertEqual(df["signal2"].iloc[-1], 0.0510)
        self.assertEqual(df["signal3"].iloc[-1], -34.5678)

        self.assertEqual(df["vout1"].iloc[0], 0.505)
        self.assertEqual(df["vout2"].iloc[0], 0.501)
        self.assertEqual(df["vout3"].iloc[0], 0.500)
        self.assertEqual(df["vout1"].iloc[-1], 0.505)
        self.assertEqual(df["vout2"].iloc[-1], -0.500)
        self.assertEqual(df["vout3"].iloc[-1], 0.499)

        self.assertTrue(df["enable1"].iloc[0])
        self.assertTrue(df["enable2"].iloc[0])
        self.assertTrue(df["enable3"].iloc[0])
        self.assertTrue(df["enable1"].iloc[-1])
        self.assertFalse(df["enable2"].iloc[-1])
        self.assertFalse(df["enable3"].iloc[-1])

    def test_load_webapp_v1v2(self):
        # web app file
        testfile = "tests/web1chan_example"
        self.parser.identify_file_source(testfile)
        self.assertEqual(self.parser.source, "WEB-APP")
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(len(df.columns), 4)

        self.assertEqual(df["vout1"].iloc[-1], 0.3)
        self.assertEqual(df["signal1"].iloc[-1], 2.77)
        self.assertTrue(df["enable1"].iloc[-1])

        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-07-26 17:12:30"))

        testfile = "tests/web2chan_example"
        self.parser.identify_file_source(testfile)
        self.assertEqual(self.parser.source, "WEB-APP")
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(len(df.columns), 7)

        self.assertEqual(df["vout1"].iloc[-1], 0.5)
        self.assertEqual(df["signal1"].iloc[-1], 5)
        self.assertTrue(df["enable1"].iloc[-1])

        self.assertEqual(df["vout2"].iloc[-1], 0.4)
        self.assertEqual(df["signal2"].iloc[-1], 15)
        self.assertTrue(df["enable2"].iloc[-1])

        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2019-07-09 17:35:24"))

        testfile = "tests/web3chan_example"
        self.parser.identify_file_source(testfile)
        self.assertEqual(self.parser.source, "WEB-APP")
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data
        self.assertEqual(len(df.columns), 10)

        self.assertEqual(df["vout1"].iloc[-1], 0.508)
        self.assertEqual(df["signal1"].iloc[-1], 2.48)
        self.assertTrue(df["enable1"].iloc[-1])

        self.assertEqual(df["vout2"].iloc[-1], -0.501)
        self.assertEqual(df["signal2"].iloc[-1], 0.03)
        self.assertTrue(df["enable2"].iloc[-1])

        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2020-02-10 11:04:54"))

    def test_load_vfp600(self):
        # vfp600
        start_time = "2019/01/01 12:00"
        testfile = "tests/vfp_example"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, False))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp(start_time))
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, True))
        df = self.parser.data
        self.assertTrue(df["timestamp"].iloc[0] < pd.Timestamp(start_time))
        ts = [df["timestamp"].iloc[-1], pd.Timestamp(start_time)]
        self.assertTrue(pd.Timedelta(max(ts) - min(ts)).seconds < 1)

        self.assertEqual(df["vout1"].iloc[-1], 0.39955)
        self.assertEqual(round(df["signal1"].iloc[-1], 2), 3.8)
        self.assertTrue(df["enable1"].iloc[-1])

        # vfp600 #2
        start_time = "2019/01/01 12:00"
        testfile = "tests/vfp_example2"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, False))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp(start_time))
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, True))
        df = self.parser.data
        self.assertTrue(df["timestamp"].iloc[0] < pd.Timestamp(start_time))
        ts = [df["timestamp"].iloc[-1], pd.Timestamp(start_time)]
        self.assertTrue(pd.Timedelta(max(ts) - min(ts)).seconds < 1)

        self.assertEqual(df["vout1"].iloc[-2], 0.101215)
        self.assertEqual(round(df["signal1"].iloc[-2], 2), 0.19)
        self.assertEqual(df["vout1"].iloc[-1], 0.101219)
        self.assertEqual(round(df["signal1"].iloc[-1], 3), 0.244)
        self.assertTrue(df["enable1"].iloc[-1])

    def test_load_explain(self):
        # gamry explain
        start_time = "2018/09/11 12:00"
        testfile = "tests/gamry_example"
        self.parser.identify_file_source(testfile)
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, False))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp(start_time))
        self.assertTrue(self.parser.load_rawfile(testfile, start_time, True))
        df = self.parser.data
        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp(start_time))

        self.assertEqual(df["vout1"].iloc[-1], 0.499654)
        self.assertEqual(df["signal1"].iloc[-1], 3.60525)
        self.assertTrue(df["enable1"].iloc[-1])

    def test_load_chinstruments(self):
        testfile = "tests/chinstruments_example"
        self.parser.identify_file_source(testfile)
        self.assertEqual(self.parser.source, "CHI")
        self.assertTrue(self.parser.load_rawfile(testfile, None, None))
        df = self.parser.data

        self.assertEqual(df["timestamp"].iloc[0], pd.Timestamp("2020/07/15 15:27:12"))
        self.assertEqual(df["timestamp"].iloc[-1], pd.Timestamp("2020/07/15 15:37:12"))

        self.assertEqual(df["vout1"].iloc[0], 0.5)
        self.assertEqual(df["vout1"].iloc[-1], 0.5)
        self.assertEqual(df["vout2"].iloc[0], -0.5)
        self.assertEqual(df["vout2"].iloc[-1], -0.5)
        self.assertEqual(df["vout3"].iloc[0], -0.3)
        self.assertEqual(df["vout3"].iloc[-1], -0.3)
        self.assertEqual(df["vout4"].iloc[0], 1.0)
        self.assertEqual(df["vout4"].iloc[-1], 1.0)

        self.assertTrue(df["enable1"].iloc[-1])
        self.assertTrue(df["enable1"].iloc[0])
        self.assertTrue(df["enable2"].iloc[-1])
        self.assertTrue(df["enable2"].iloc[0])
        self.assertTrue(df["enable3"].iloc[-1])
        self.assertTrue(df["enable3"].iloc[0])
        self.assertTrue(df["enable4"].iloc[-1])
        self.assertTrue(df["enable4"].iloc[0])

        self.assertEqual(df["signal1"].iloc[-1], 0.3608)
        self.assertEqual(df["signal2"].iloc[-1], 20.09)
        self.assertEqual(df["signal3"].iloc[-1], 8)
        self.assertEqual(df["signal4"].iloc[-1], 39.8)



import numpy as np
import scipy.stats
import scipy.optimize
import pandas as pd
import re
import copy
from datetime import datetime, timedelta

# logging
import logging

log = logging.getLogger()


class PSenseCalibration(object):
    "Compute the sensitivity, background signal of a sensor, and its correlation with reference values"

    def __init__(
        self, expid, data, sample_offset_time=-5, sample_window_time=5, debugmode=False
    ):
        assert isinstance(data, pd.DataFrame), "data must be a pandas DataFrame"
        self.debug = debugmode
        self.msgid = "[PSenseCalibration]"
        self.expid = expid

        self.raw = data
        r = re.compile("signal*")
        self.channels = list(filter(r.match, data.columns.values.tolist()))

        self.calibration = pd.DataFrame()
        self.sample_offset = timedelta(minutes=sample_offset_time)
        self.sample_window = timedelta(minutes=sample_window_time)

        self.result = dict(coeff=[], r=[])

    def get_sensordata_slice(self, t_start, t_finish):
        log.debug("data from {} to {}".format(t_start, t_finish))
        mask = (self.raw.index >= t_start) & (self.raw.index < t_finish)
        return self.raw.loc[mask]

    def add_reference(self, ts, reference):
        log.info("ts={}".format(ts.isoformat()))
        # replace existing calibration point if the timestamp is the same
        mask = self.calibration.index == ts
        if len(self.calibration.loc[mask]) > 0:
            self.calibration = self.calibration.loc[~mask]

        sensor = self.get_sensordata_slice(
            ts + self.sample_offset, ts + self.sample_offset + self.sample_window
        )

        record = [ts.replace(tzinfo=None), reference]
        keys = ["ts", "ref"]
        for channel in self.channels:
            record.append(np.round(np.mean(sensor[channel]), 2))
            keys.append(channel)

        log.debug("record={}".format(record[1:]))

        record = pd.DataFrame([record], columns=keys)
        if self.calibration.count == 0:
            record.set_index("ts")
            self.calibration = record
        else:
            self.calibration = self.calibration.append(record)

        return True

    def dump(self):
        return copy.deepcopy(self.calibration)

    def model_fit(self, model_type="linear"):
        assert (
            len(self.calibration) > 1
        ), "model_fit requires more than 1 calibration point"

        # numpy doesn't handle NaN values properly, we need to drop from the dataframe prior to analysis
        calibration = self.calibration.dropna(axis=0, how="any", inplace=False)

        # per each sensor
        reference = calibration["ref"]
        for channel in self.channels:
            sensor = calibration[channel]

            if model_type == "linear":
                slope, intercept, r, p, stderr = scipy.stats.linregress(
                    reference, sensor
                )
                self.result["coeff"].append([round(slope, 4), round(intercept, 2)])
                self.result["r"].append(round(r, 4))
                log.debug("{} fit: y={}x+{}".format(model_type, slope, intercept))

            elif model_type == "logarithmic":

                def exp_func(x, a, b, c):
                    return a * np.exp(-b * x) + c

                popt, pcov = scipy.optimize.curve_fit(exp_func, reference, sensor)
                self.result["coeff"].append(popt)

                ss_res = np.sum(
                    (sensor - exp_func(reference, *popt)) ** 2
                )  # sum of residuals
                ss_tot = np.sum((sensor - np.mean(sensor)) ** 2)  # total sum of squares
                r2 = 1 - (ss_res / ss_tot)  # r-squared
                self.result["r"].append(round(np.sqrt(r2), 4))  # pearson
                log.debug(
                    "{} fit: y= {}*exp(-{}*x) + {}".format(
                        model_type, popt[0], popt[1], popt[2]
                    )
                )

        return copy.deepcopy(self.result)



import time
import pytz
from datetime import datetime, timedelta
import pandas as pd

# dynamodb
import boto3
from botocore.client import Config
from botocore.exceptions import ReadTimeoutError
from boto3.dynamodb.conditions import Key, Attr
from dynamodb_json import json_util
from decimal import Decimal
import logging
import types

# percusense
from psense_common import psense_format

valid_types = [
    "Glucose",
    "Lactate",
    "Oxygen",
    "Ketone",
    "Temperature",
    "Multianalyte",
    "Choline",
    "Acetylcholine",
    "Interferent",
]

# logging
log = logging.getLogger()


""" functions """


def is_float(val):
    "check if the input value is a floating point number"
    try:
        # if this isn't a numeric value, we'll cause a valueerror exception
        float(val)
        # don't allow NaN values, either.
        if pd.isnull(val):
            return False
        return True
    except BaseException:
        return False


# override the _flush() method of Table.batch_writer(). This modification just persists the response from AWS to self._response
def _flush(self):
    items_to_send = self._items_buffer[: self._flush_amount]
    self._items_buffer = self._items_buffer[self._flush_amount :]
    self._response = self._client.batch_write_item(
        RequestItems={self._table_name: items_to_send}
    )
    unprocessed_items = self._response["UnprocessedItems"]

    if unprocessed_items and unprocessed_items[self._table_name]:
        # Any unprocessed_items are immediately added to the
        # next batch we send.
        self._items_buffer.extend(unprocessed_items[self._table_name])
    else:
        self._items_buffer = []

    log.debug(
        "Batch write sent %s, unprocessed: %s",
        len(items_to_send),
        len(self._items_buffer),
    )


""" classes """


class PSenseAWSInterface(object):
    def __init__(self, profile_name="pshield", debugmode=False, testmode=False):
        self.msgid = "[PSenseAWSInterface]"
        self.debug = debugmode
        self.test_mode = testmode
        self.batch_req_limit = 25
        self.batch_req_cooldown_secs = 10
        self.query_req_limit = 5000
        self.query_req_count_limit = 2

        self.local = pytz.timezone("America/Los_Angeles")
        self.vset = {
            "default": 0,
            "Glucose": 0.5,
            "Lactate": 0.5,
            "Oxygen": -0.5,
            "Ketone": 0.6,
        }
        self.we_count = 1

        # initialize DynamoDB connection
        config = Config(connect_timeout=5, read_timeout=5, retries={"max_attempts": 0})
        try:
            self.aws = boto3.session.Session(profile_name=profile_name)
        except BaseException:
            # profile doesn't exist
            log.debug(
                "failed to load profile: {}. Loading defaults".format(profile_name)
            )
            self.aws = boto3.session.Session(region_name="us-east-1")

        self.db = self.aws.resource("dynamodb", config=config)
        self.table_experiments = self.db.Table("Experiments")
        self.table_sensordata = self.db.Table("SensorData")
        self.table_eventdata = self.db.Table("CalibrationData")

        # check for validity
        self.pf = psense_format.psense_exp_title()

        # initialize experiment information
        self.experiment_name = ""
        self.experiment_info = []
        self.exp_is_valid = False

    def set_query_config(self, req_size=5000, req_count=2):
        """change dynamodb query parameters based on size:
        req_size: maximum number of items to return
        req_count: number of queries to allow (total = req_size*req_count)
        """
        self.query_req_limit = req_size
        self.query_req_count_limit = req_count
        log.debug("query_req limits -- count: {} size: {}".format(req_count, req_size))

    def get_experiment(self):
        "return the experiment id"
        return self.experiment_name

    def set_experiment(self, experiment_name):
        "update the experiment id"
        self.experiment_name = experiment_name

    def verif_experiment(self, experiment_name=""):
        "return whether the experiment id follows the psense naming convention"
        if len(experiment_name) > 0:
            self.set_experiment(experiment_name)

        experiment_name = self.get_experiment()
        self.experiment_info = self.pf.decode(experiment_name)

        try:
            # invalid experiment name, or missing data
            # assert len(experiment_name) == 32
            assert isinstance(self.experiment_info, dict)
            assert isinstance(self.experiment_info["study"], dict)
            assert isinstance(self.experiment_info["sensor"], dict)

            assert self.experiment_info["study"]["type"]
            assert self.experiment_info["study"]["run"].isnumeric()
            assert len(self.experiment_info["study"]["device"]) > 2

            sensor_code = self.experiment_info["sensor"]["code"].split("-")
            assert len(sensor_code) == 3
            assert len(sensor_code[0]) == 4
            assert len(sensor_code[1]) == 3
            assert len(sensor_code[2]) == 2

            assert len(self.experiment_info["sensor"]["vsn"]) == 2  # 2-digit vsn

            assert isinstance(self.experiment_info["sensor"]["electrodes"], list)
            we_count = self.experiment_info["sensor"]["count"]
            if we_count.isnumeric():
                assert len(self.experiment_info["sensor"]["electrodes"]) == int(
                    we_count
                )
                assert len(experiment_name) == 25 + len(
                    self.experiment_info["study"]["device"]
                ) + 3 * int(
                    we_count
                )  # three letters per electrode, no more or less
            else:
                assert len(self.experiment_info["sensor"]["electrodes"]) == 1
                assert len(experiment_name) == 29 + len(
                    self.experiment_info["study"]["device"]
                )  # 32 letters in the old experiment id format

            assert isinstance(self.experiment_info["sensor"]["electrodes"][0], dict)

        except BaseException as E:
            log.info("[{}] invalid: {}".format(experiment_name, E))
            self.exp_is_valid = False
            return False

        log.info("[{}] valid".format(experiment_name))
        self.exp_is_valid = True
        return True

    def check_experiment_data(self, experiment_name):
        "return whether an experiment exists and has data"
        log.debug("Checking for items associated with {}".format(experiment_name))

        if self.test_mode:
            return ("Yes", True, -999)

        res = "No"
        is_active = False
        last_vout = "0"

        if len(experiment_name) > 0:  # skip if there is no data
            try:
                response = self.table_sensordata.query(
                    KeyConditionExpression=Key("experiment").eq(experiment_name),
                    Limit=1,  # +return a single data point
                    ScanIndexForward=False,  # +return the most recent value
                )
            except ReadTimeoutError:
                log.error(
                    "failed. Timeout on query for experiment={}".format(experiment_name)
                )
                return (None, None, None)

            if response["Count"] > 0:
                res = "Yes"
                last_received = (
                    datetime.utcnow()
                    - datetime.fromisoformat(response["Items"][0]["timestamp"][:19])
                ).total_seconds()
                is_active = last_received < 3600

                last_enable = [
                    value
                    for key, value in sorted(response["Items"][0].items())
                    if "enable" in key
                ]
                last_vout = [
                    value
                    for key, value in sorted(response["Items"][0].items())
                    if "vout" in key
                ]
                if len(last_enable):  # are there enable flags in the data?
                    last_vout = " / ".join(
                        [
                            "{:.3f}".format(last_vout[i]) if enabled else "(off)"
                            for i, enabled in enumerate(last_enable)
                        ]
                    )
                else:
                    last_vout = " / ".join(map("{:.3f}".format, last_vout))

        return (res, is_active, last_vout)

    def add_experiment(self, device_version="unknown", sample_period=30, vset=None):
        "Push experiment id to the Experiments table"
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)

        if self.test_mode:
            response = dict(Items=[])
        else:
            try:
                response = self.table_experiments.query(
                    KeyConditionExpression=Key("experiment").eq(
                        expid
                    ),  # check for a single experiment id
                    Limit=1,  # +return a single data point
                )
            except ReadTimeoutError:
                log.error("failed. Timeout on query for experiment={}".format(expid))
                return False, "timeout"

        if vset is None:
            vset = self.vset.get(
                self.experiment_info["sensor"]["electrodes"][0]["type"],
                self.vset["default"],
            )

        if len(response["Items"]) > 0:
            log.error("failed. [{}] already exists.".format(expid))
            return False, "exists"
        else:
            item = {
                "experiment": expid,
                "t_start": datetime.utcnow().isoformat(),
                "auto_creation": 1,
                "dev_id": "{}".format(self.experiment_info["study"]["device"]),
                "dev_ver": device_version,
                "we_count": "{}".format(self.we_count),
                "we_types": self.experiment_info["sensor"]["electrodes"],
                "v_out": "{}".format(vset),
                "t_samp": "{}".format(sample_period),
            }
            if not self.test_mode:
                try:
                    self.table_experiments.put_item(Item=item)
                except BaseException:
                    log.error("failed. [{}] could not be added.".format(expid))
                    return False, "error"

        log.info("success. [{}] added to dynamodb".format(expid))
        return True, ""

    def add_sensordata(self, timestamp=[], vout=None, signal=[0], enable=True):
        "Push (time,vout,isig,enable) tuple to the SensorData table"
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(
            self.get_experiment()
        )

        if isinstance(timestamp, datetime):
            if timestamp.tzinfo is None:
                timestamp = self.local.localize(timestamp).astimezone(pytz.utc)
            else:
                timestamp = timestamp.astimezone(pytz.utc)
        elif isinstance(timestamp, str):
            timestamp = self.local.localize(
                datetime.fromisoformat(timestamp)
            ).astimezone(pytz.utc)
        else:
            timestamp = datetime.utcnow()

        if vout is None:
            vout = [
                self.vset.get(
                    self.experiment_info["sensor"]["electrodes"][0]["type"],
                    self.vset["default"],
                )
            ] * len(signal)

        if is_float(signal):
            signal = [signal]
        if is_float(vout):
            vout = [vout] * len(signal)
        if is_float(enable):
            enable = [enable] * len(signal)

        item = {
            "experiment": self.get_experiment(),
            # timestamps are UTC. THe web app plots it is in pacific
            "timestamp": timestamp.replace(tzinfo=None).isoformat(),
        }
        for i in range(len(signal)):
            item["signal{}".format(i + 1)] = Decimal(
                str(round(signal[i], 2))
            )  # floats -> Decimal for ddb
            item["vout{}".format(i + 1)] = Decimal(
                str(round(vout[i], 3))
            )  # floats -> Decimal for ddb
            item["enable{}".format(i + 1)] = enable[i]  # booleans are supported in ddb

        if self.test_mode:
            log.info("test mode put_item successful")
            return True

        try:
            self.table_sensordata.put_item(Item=item)
            log.info("sensor data put_item was successful")
        except BaseException:
            log.error("sensor data put_item failed. {}".format(item))
            return False

        return True

    def generate_put_request(self, row):
        "helper function for batch writing sensor data"
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(
            self.get_experiment()
        )

        # extract signals from the provided row
        request = {
            "experiment": self.get_experiment(),
            "timestamp": row[0],
        }
        for i in range(int((len(row) - 1) / 3)):
            request["signal{}".format(i + 1)] = Decimal(str(round(row[3 * i + 2], 2)))
            request["vout{}".format(i + 1)] = Decimal(str(round(row[3 * i + 1], 3)))
            request["enable{}".format(i + 1)] = row[3 * i + 3]

        # build the request dict
        return request

    def add_bulk_sensordata(self, data):
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(
            self.get_experiment()
        )

        log.info(
            "add_bulk_sensordata: batch writing {} records to SensorData".format(
                len(data)
            )
        )

        requests = []
        for sample in data:
            timestamp = sample[0]
            if isinstance(timestamp, str):
                timestamp = pytz.utc.localize(
                    datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f")
                )
            elif isinstance(timestamp, pd.Timestamp):
                timestamp = self.local.localize(timestamp.to_pydatetime()).astimezone(
                    pytz.utc
                )
            else:
                timestamp = (timestamp).astimezone(pytz.utc)

            sample[0] = timestamp.replace(tzinfo=None).isoformat()
            requests.append(self.generate_put_request(sample))

            if len(requests) >= self.batch_req_limit:
                log.debug(
                    "add_bulk_sensordata: inserting batch of [25] records into [SensorData]"
                )

                if not self.test_mode:
                    with self.table_sensordata.batch_writer(
                        overwrite_by_pkeys=["experiment", "timestamp"]
                    ) as batch:
                        batch._flush = types.MethodType(_flush, batch)
                        for request in requests:
                            batch.put_item(Item=request)

                    res = batch._response
                    if res["ResponseMetadata"].get("HTTPStatusCode", None) != 200:
                        res = batch._response
                        log.warning(
                            "add_bulk_sensordata: batch_writer() returned unexpected response: [RequestId] {} [Status] {} [RetryAttempts] {} [Unprocessed] {}".format(
                                res["ResponseMetadata"].get("RequestId", None),
                                res["ResponseMetadata"].get("HTTPStatusCode", None),
                                res["ResponseMetadata"].get("RetryAttempts", None),
                                res.get("UnprocessedItems", None),
                            )
                        )
                        log.warning(
                            "add_bulk_sensordata: batch_writer() batch = {}".format(
                                requests
                            )
                        )
                        return False

                    time.sleep(self.batch_req_cooldown_secs)

                requests = []

        if not self.test_mode and len(requests) > 0:
            log.debug(
                "add_bulk_sensordata: inserting batch of [{}] records into [SensorData]".format(
                    len(requests)
                )
            )

            with self.table_sensordata.batch_writer(
                overwrite_by_pkeys=["experiment", "timestamp"]
            ) as batch:
                batch._flush = types.MethodType(_flush, batch)
                for request in requests:
                    batch.put_item(Item=request)

            res = batch._response
            if res["ResponseMetadata"].get("HTTPStatusCode", None) != 200:
                log.warning(
                    "add_bulk_sensordata: batch_writer() returned unexpected response: [RequestId] {} [Status] {} [RetryAttempts] {} [Unprocessed] {}".format(
                        res["ResponseMetadata"].get("RequestId", None),
                        res["ResponseMetadata"].get("HTTPStatusCode", None),
                        res["ResponseMetadata"].get("RetryAttempts", None),
                        res.get("UnprocessedItems", None),
                    )
                )
                log.warning(
                    "add_bulk_sensordata: batch_writer() batch = {}".format(requests)
                )
                return False

        return True

    def get_experiments(self):
        "request a list of sensor experiments from aws"
        try:
            response = self.table_experiments.query(
                IndexName="latest-experiments-index",
                KeyConditionExpression=Key("auto_creation").eq(1),
                Limit=self.query_req_limit,
                ScanIndexForward=False,
            )
        except ReadTimeoutError:
            log.error("failed. Timeout on experiments table table")
            return []

        data = json_util.loads(response)
        records = data["Items"]
        experiments = set()
        for exp in records:
            experiments.add(exp["experiment"])

        log.debug("returned {} unique experiment ids".format(len(experiments)))
        return list(sorted(experiments))

    def get_sensor_data(self):
        "request sensor data from aws"
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)

        if self.test_mode:
            res = {
                "Count": 1,
                "Items": [
                    {
                        "experiment": expid,
                        "timestamp": (
                            datetime.utcnow() - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 10,
                        "vout1": 0.4,
                        "enable1": True,
                    }
                ],
            }
        else:
            try:
                res = self.table_sensordata.query(
                    KeyConditionExpression=Key("experiment").eq(expid),
                    Limit=self.query_req_limit,
                )
            except ReadTimeoutError:
                log.error("failed. Timeout on query for experiment={}".format(expid))
                return None, None

        count, res = self.aws_to_df(res)
        log.info("[{}] initial request found {} records".format(expid, count))

        more_to_retrieve = count == self.query_req_limit
        num_queries_executed = 1
        while more_to_retrieve and num_queries_executed < self.query_req_count_limit:
            last_received_timestamp = datetime.isoformat(res.index[-1])
            q_count, q_res = self.get_sensordata_sincetimestamp(
                last_received_timestamp, num_queries_executed
            )

            res = res.append(q_res)
            count += q_count

            num_queries_executed += 1
            more_to_retrieve = q_count % self.query_req_limit == 0

        log.info("[{}] found {} records".format(expid, count))

        return count, res

    def get_sensordata_sincetimestamp(self, last_received_timestamp, query_count=0):
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)

        log.info(
            "[{}] [count={}] (start={})".format(
                expid, query_count, last_received_timestamp
            )
        )

        if self.test_mode:
            res = {
                "Count": 1,
                "Items": [
                    {
                        "experiment": expid,
                        "timestamp": (
                            datetime.utcnow() - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 10,
                        "vout1": 0.4,
                        "enable1": True,
                    }
                ],
            }
        else:
            try:
                res = self.table_sensordata.query(
                    KeyConditionExpression=Key("experiment").eq(expid)
                    & Key("timestamp").gt(last_received_timestamp),
                    Limit=self.query_req_limit,
                )
            except ReadTimeoutError:
                log.error("failed. Timeout on query for experiment={}".format(expid))
                return None, None

        count, res = self.aws_to_df(res)

        query_count += 1
        log.debug("[{}] returned {} items".format(expid, count))

        # recursively query for more data based on last timestamp
        if count == self.query_req_limit and query_count < self.query_req_count_limit:
            last_received_timestamp = datetime.isoformat(res.index[-1])

            q_count, q_result = self.get_sensordata_sincetimestamp(
                last_received_timestamp, query_count
            )

            res = res.append(q_result)
            count += q_count

        return count, res

    def get_sensordata_slice(self, timestamp_start, samples):
        "get x samples of data starting from a specific timestamp"
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)

        if self.test_mode:
            res = {
                "Count": 1,
                "Items": [
                    {
                        "experiment": expid,
                        "timestamp": (
                            datetime.utcnow() - timedelta(minutes=10)
                        ).isoformat("T"),
                        "signal1": 5,
                        "vout1": 0.2,
                        "enable1": True,
                    }
                ],
            }
        else:
            try:
                res = self.table_sensordata.query(
                    KeyConditionExpression=Key("experiment").eq(expid)
                    & Key("timestamp").gt(timestamp_start),
                    Limit=samples,
                )
            except ReadTimeoutError:
                log.error("failed. Timeout on query for experiment={}".format(expid))
                return None, None

        count, res = self.aws_to_df(res)

        log.debug("[{}] ({}) returned {} items".format(expid, timestamp_start, count))
        return count, res

    def aws_to_df(self, response):
        "convert a aws response to a pandas dataframe, indexed by timestamp"

        data = json_util.loads(response)
        records = data["Items"]
        count = data["Count"]

        res = pd.json_normalize(records)
        if count > 0:  # only attempt modification if there are records
            res.drop(columns="experiment", inplace=True)
            try:
                res["timestamp"] = pd.to_datetime(res["timestamp"])
            except BaseException:
                pass

            try:
                res["timestamp"] = res["timestamp"].apply(
                    lambda x: x.replace(microsecond=0)
                )  # trim microseconds
            except BaseException:
                pass
            res.set_index("timestamp", inplace=True)

        return count, res

    def get_event_data(self):
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)

        if self.test_mode:
            res = {
                "Count": 1,
                "Items": [
                    {
                        "experiment": expid,
                        "timestamp": "1|{}".format(
                            (datetime.utcnow() - timedelta(minutes=10)).isoformat("T")
                        ),
                        "reference": 100,
                        "sensor": 10,
                        "type": "Glucose",
                    }
                ],
            }
        else:
            try:
                res = self.table_eventdata.query(
                    KeyConditionExpression=Key("experiment").eq(expid)
                )
            except ReadTimeoutError:
                log.error("failed. Timeout on query for experiment={}".format(expid))
                return None

        data = json_util.loads(res)
        records = data["Items"]
        count = data["Count"]

        res = pd.json_normalize(records)
        res["set"] = res["timestamp"].map(lambda a: a.split("|")[0])
        res["timestamp"] = res["timestamp"].map(lambda a: a.split("|")[1])
        log.debug("found {} items".format(count))

        return res

    def add_event(self, set, timestamp, type, reference, sensor):
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)
        assert type in valid_types, "Type [{}] is not valid".format(type)
        try:
            dt = pytz.utc.localize(datetime.fromisoformat(timestamp))
        except BaseException:
            dt = timestamp.astimezone(pytz.utc)

        log.debug("insert @ {}".format(dt.isoformat()))
        newitem = {
            "experiment": expid,
            "timestamp": "{}|{}".format(set, dt.replace(tzinfo=None).isoformat()),
            "type": type,
            "reference": Decimal(str(reference)),
            "sensor": Decimal(str(sensor)),
        }

        if self.test_mode:
            return True

        response = self.table_eventdata.put_item(Item=newitem)
        log.debug("response: {}".format(response))

    def delete_event(self, set, timestamp, type):
        expid = self.get_experiment()
        assert self.exp_is_valid, "Experiment [{}] has not been validated".format(expid)
        assert type in valid_types, "Type [{}] is not valid".format(type)
        try:
            dt = pytz.utc.localize(datetime.fromisoformat(timestamp))
        except BaseException:
            dt = timestamp.astimezone(pytz.utc)

        log.debug("{} ({}, {})".format(dt.replace(tzinfo=None).isoformat(), type, set))

        if self.test_mode:
            return True

        response = self.table_eventdata.delete_item(
            Key={
                "experiment": expid,
                "timestamp": "{}|{}".format(set, dt.replace(tzinfo=None).isoformat()),
            }
        )
        response = json_util.loads(response)
        return response



import unittest
from unittest.mock import patch
from psense_common import psense_format
from datetime import datetime


class TestPFormat(unittest.TestCase):
    def setUp(self):
        self.format = psense_format.psense_exp_title()

    def test_invalid_input(self):
        self.assertEqual(self.format.decode(""), {"study": [], "sensor": []})

    def test_valid_input_legacy(self):
        code1 = "19C12B904B02-1111-222-33-O02CPBX"
        self.assertEqual(
            self.format.decode(code1),
            {
                "study": {
                    "start": "2019-03-12",
                    "type": "Beaker",
                    "run": "04",
                    "device": "B02",
                },
                "sensor": {
                    "code": "1111-222-33",
                    "count": "O",
                    "vsn": "02",
                    "electrodes": [
                        {
                            "type": "Oxygen",
                            "electrode": "Circular",
                            "stack": "Planar",
                            "process": "Blanket",
                            "extras": "X",
                        }
                    ],
                },
            },
        )

        code2 = "19C11A001B04-0205-205-LA-G11CPVX"
        self.assertEqual(
            self.format.decode(code2),
            {
                "study": {
                    "start": "2019-03-11",
                    "type": "Heat Block",
                    "run": "01",
                    "device": "B04",
                },
                "sensor": {
                    "code": "0205-205-LA",
                    "count": "G",
                    "vsn": "11",
                    "electrodes": [
                        {
                            "type": "Glucose",
                            "electrode": "Circular",
                            "stack": "Planar",
                            "process": "Volcano",
                            "extras": "X",
                        }
                    ],
                },
            },
        )

        code3 = "19H01P101B08-500A-123-LA-X13RPSN"
        self.assertEqual(
            self.format.decode(code3),
            {
                "study": {
                    "start": "2019-08-01",
                    "type": "Prototype",
                    "run": "01",
                    "device": "B08",
                },
                "sensor": {
                    "code": "500A-123-LA",
                    "count": "X",
                    "vsn": "13",
                    "electrodes": [
                        {
                            "type": "Multi (legacy)",
                            "electrode": "Rectangular",
                            "stack": "Planar",
                            "process": "Standard",
                            "extras": "NAD+",
                        }
                    ],
                },
            },
        )

    def test_valid_input_multianalyte(self):
        code1 = "19C12B904B02-1111-222-33-199OCB"
        # new format, but just 1 electrode
        self.assertEqual(
            self.format.decode(code1),
            {
                "study": {
                    "start": "2019-03-12",
                    "type": "Beaker",
                    "run": "04",
                    "device": "B02",
                },
                "sensor": {
                    "code": "1111-222-33",
                    "count": "1",
                    "vsn": "99",
                    "electrodes": [
                        {
                            "type": "Oxygen",
                            "electrode": "Circular",
                            "process": "Blanket",
                        }
                    ],
                },
            },
        )

        # dual analyte
        code2 = "20G11AN07B04-0205-205-LA-215GCVOCS"
        self.assertEqual(
            self.format.decode(code2),
            {
                "study": {
                    "start": "2020-07-11",
                    "type": "Pre-Clinical",
                    "run": "07",
                    "device": "B04",
                },
                "sensor": {
                    "code": "0205-205-LA",
                    "count": "2",
                    "vsn": "15",
                    "electrodes": [
                        {
                            "type": "Glucose",
                            "electrode": "Circular",
                            "process": "Volcano",
                        },
                        {
                            "type": "Oxygen",
                            "electrode": "Circular",
                            "process": "Standard",
                        },
                    ],
                },
            },
        )

    def test_partial_input(self):
        code_study_only = "19A01A001B04"
        self.assertEqual(
            self.format.decode(code_study_only),
            {
                "study": {
                    "start": "2019-01-01",
                    "type": "Heat Block",
                    "run": "01",
                    "device": "B04",
                },
                "sensor": [],
            },
        )

        code_partial = "19C11JJXXM99-WE-RE-X-X99CXXX"
        self.assertEqual(
            self.format.decode(code_partial),
            {
                "study": {
                    "start": "2019-03-11",
                    "type": None,
                    "run": "XX",
                    "device": "M99",
                },
                "sensor": {
                    "code": "WE-RE-X",
                    "count": "X",
                    "vsn": "99",
                    "electrodes": [
                        {
                            "type": "Multi (legacy)",
                            "electrode": "Circular",
                            "stack": "Other",
                            "process": None,
                            "extras": "X",
                        }
                    ],
                },
            },
        )

    def test_yn_input(self):
        with patch("builtins.input", return_value=""):
            self.assertTrue(psense_format.yn_input("message"))
            self.assertFalse(psense_format.yn_input("message", "n"))

        with patch("builtins.input", return_value="y"):
            self.assertTrue(psense_format.yn_input("message"))
            self.assertTrue(psense_format.yn_input("message", "n"))

    def test_answer_input(self):
        with patch("builtins.input", return_value=""):
            self.assertEqual(
                psense_format.answer_input(
                    "message", "[0-9]", default="0", is_required=False
                ),
                "0",
            )

        with patch("builtins.input", return_value="1"):
            self.assertEqual(
                psense_format.answer_input(
                    "message", "[0-9]", default="0", is_required=False
                ),
                "1",
            )
            self.assertEqual(
                psense_format.answer_input(
                    "message", ["0", "1", "2"], default="0", is_required=False
                ),
                "1",
            )

    def test_exptype_input(self):
        with patch("builtins.input", return_value=""):
            self.assertTrue(psense_format.exptype_input())

        with patch("builtins.input", return_value="1"):
            self.assertTrue(psense_format.exptype_input())

        with patch("builtins.input", return_value="2"):
            self.assertFalse(psense_format.exptype_input())

    def test_setup_experiment(self):
        self.assertRaises(AssertionError, psense_format.setup_new_experiment, 0)

        datecode = "{}{}{}".format(
            datetime.now().strftime("%y"),
            chr(int(datetime.now().strftime("%m")) + 64),
            datetime.now().strftime("%d"),
        )

        with patch(
            "builtins.input",
            side_effect=["A1", "", "9999", "", "", "99", "1", "K", "", "S"],
        ):
            res = psense_format.setup_new_experiment("B01")
            self.assertEqual(res[5:], "A101B01-9999-000-00-199KCS")
            self.assertEqual(res[:5], datecode)

        with patch(
            "builtins.input",
            side_effect=[
                "AN",
                "5",
                "9999",
                "",
                "",
                "99",
                "3",
                "G",
                "C",
                "V",
                "L",
                "R",
                "V",
                "A",
                "",
                "W",
            ],
        ):
            res = psense_format.setup_new_experiment("B21")
            self.assertEqual(res[5:], "AN05B21-9999-000-00-399GCVLRVACW")
            self.assertEqual(res[:5], datecode)

        with patch(
            "builtins.input",
            side_effect=[
                "M1",
                "B1",
                "A",
                "99",
                "8888",
                "777",
                "66",
                "L",
                "88",
                "L",
                "1",
                "L",
                "AAA",
                "",
                "",
                "B",
            ],
        ):
            res = psense_format.setup_new_experiment("F01")
            self.assertEqual(res[5:], "B199F01-8888-777-66-188LCB")


if __name__ == "__main__":
    unittest.main()



import unittest
import pandas as pd
import numpy as np
import json
from psense_common.analysis import StepResponse, PSenseResponse
from datetime import datetime


class TestStepResponse(unittest.TestCase):
    def setUp(self):
        self.test_sensor = pd.read_csv(
            "tests/testdata_stepresponse",
            sep=",",
            index_col="timestamp",
            parse_dates=["timestamp"],
        )
        self.stepresponse = StepResponse(data=self.test_sensor)
        self.expected_result = {
            "signal1": {
                "max_achieved_rate": 0.62,
                "response_time": 61.6,
                "t10_ts": "2019-04-02T09:05:11",
                "t10_val": 3.22,
                "t90_ts": "2019-04-02T10:06:49",
                "t90_val": 8.55,
            }
        }
        self.presponse = PSenseResponse(df=self.test_sensor)

    def test_sr_init(self):
        self.assertRaises(AssertionError, StepResponse, [0, 1, 2])

        sr = StepResponse(data=self.test_sensor)
        self.assertEqual(sr.to_per_minute, 2)
        self.assertEqual(len(sr.data), len(self.test_sensor))

    def test_psenseresponse_init(self):
        self.assertEqual(self.presponse.coeff_thresh_low, 0.012)
        self.assertEqual(self.presponse.coeff_thresh_high, 2)
        self.assertEqual(self.presponse.coeff_thresh_ignore_peak, 0.25)
        self.assertEqual(len(self.presponse.data), len(self.test_sensor))

    def test_sr_tostr(self):
        stringified = json.dumps(self.expected_result, indent=2)
        self.assertEqual(str(self.stepresponse), stringified)

    def test_sr_compute(self):
        stepresponse = StepResponse(
            data=self.test_sensor[: pd.Timestamp("2019-04-02 09:35:02")]
        )
        self.assertTrue(stepresponse.is_valid())
        output = stepresponse.output
        self.assertEqual(len(output), 1)
        output = output.iloc[0]
        self.assertEqual(output.T_Start, pd.Timestamp("2019-04-02 08:00:03"))
        self.assertEqual(output.Width_P, 95)
        self.assertEqual(output.I_P0, 2.65)
        self.assertEqual(output.I_P1, 4.83)
        self.assertEqual(output.Range_I_RT, 2.07)
        self.assertEqual(output.T_RT, 5.6)
        self.assertEqual(output.T_RT10, pd.Timestamp("2019-04-02 09:04:11"))
        self.assertEqual(output.I_RT10, 2.92)
        self.assertEqual(output.T_RT90, pd.Timestamp("2019-04-02 09:09:47"))
        self.assertEqual(output.I_RT90, 4.61)
        self.assertEqual(output.Max_dI, 0.34)

        stepresponse = StepResponse(
            data=self.test_sensor[: pd.Timestamp("2019-04-02 08:00:33")]
        )
        self.assertFalse(stepresponse.is_valid())

    def test_find_steps(self):

        tm_start, tm_finish = self.presponse.find_steps_in_signal(
            signal=self.test_sensor["c_signal1"], numsteps=3, samplewindow=8
        )
        self.assertEqual(len(tm_start), 3)
        self.assertEqual(
            tm_start,
            [
                pd.Timestamp("2019-04-02 09:03:41"),
                pd.Timestamp("2019-04-02 09:36:32"),
                pd.Timestamp("2019-04-02 10:04:19"),
            ],
        )
        self.assertEqual(
            tm_finish,
            [
                pd.Timestamp("2019-04-02 09:14:18"),
                pd.Timestamp("2019-04-02 09:47:10"),
                pd.Timestamp("2019-04-02 10:14:56"),
            ],
        )

        tm_start, tm_finish = self.presponse.find_steps_in_signal(
            signal=self.test_sensor["c_signal1"], numsteps=5, samplewindow=5
        )
        self.assertEqual(len(tm_start), 3)
        self.assertEqual(tm_start[0], pd.Timestamp("2019-04-02 09:03:41"))
        self.assertEqual(tm_finish[0], pd.Timestamp("2019-04-02 09:12:47"))

    def test_define_steps(self):
        self.assertRaises(
            AssertionError,
            self.presponse.define_steps_in_signal,
            "not-a-list-or-series",
        )
        injections = ["2019-04-02 09:03", "2019-04-02 09:36", "2019-04-02 10:03"]

        # check that overlapping steps aren't allowed
        self.assertRaises(
            AssertionError, self.presponse.define_steps_in_signal, injections, 30
        )

        tm_start, tm_finish = self.presponse.define_steps_in_signal(
            injections, time_window_mins=15
        )
        self.assertTrue(isinstance(tm_start, pd.Series))
        self.assertTrue(isinstance(tm_finish, pd.Series))
        self.assertTrue(len(tm_start), 3)
        self.assertTrue(len(tm_finish), 3)
        self.assertEqual(tm_start[0], pd.Timestamp(injections[0]))
        self.assertEqual(tm_finish[0], pd.Timestamp("2019-04-02 09:18:00"))
        self.assertEqual(tm_start[2], pd.Timestamp(injections[2]))
        self.assertEqual(
            tm_finish.to_list(),
            [
                pd.Timestamp("2019-04-02 09:18:00"),
                pd.Timestamp("2019-04-02 09:51:00"),
                pd.Timestamp("2019-04-02 10:18:00"),
            ],
        )

    def test_sr_from_computed_steps(self):
        tm_start, tm_finish = self.presponse.find_steps_in_signal(
            signal=self.test_sensor["c_signal1"], numsteps=3, samplewindow=8
        )

        window = self.test_sensor[
            (self.test_sensor.index >= tm_start[0])
            & (self.test_sensor.index < tm_finish[0])
        ]
        sr = StepResponse(
            data=window,
            c_thresh_ignore=self.presponse.coeff_thresh_ignore_peak,
            channels=["signal1"],
        )
        expected_output = dict(
            signal1={
                "max_achieved_rate": 0.34,
                "response_time": 4.6,
                "t10_ts": "2019-04-02T09:04:11",
                "t10_val": 2.92,
                "t90_ts": "2019-04-02T09:08:47",
                "t90_val": 4.55,
            }
        )
        self.assertEqual(str(sr), json.dumps(expected_output, indent=2))

    def test_psenseresponse_run(self):
        self.presponse.run()

        # test deepcopy
        result = self.presponse.getResult()
        result["signal1"] = result["signal1"][:1]
        self.assertEqual(len(result["signal1"]), 1)

        # test run results
        result = self.presponse.result["signal1"]
        self.assertEqual(len(result), 3)
        self.assertTrue(result.Width_P.equals(pd.Series([10.1, 10.0, 10.1])))
        self.assertTrue(result.I_P0.equals(pd.Series([2.78, 4.80, 6.73])))
        self.assertTrue(result.I_P1.equals(pd.Series([4.74, 6.62, 8.99])))
        self.assertTrue(result.T_RT.equals(pd.Series([4.6, 5.5, 3.5])))
        self.assertTrue(result.Max_dI.equals(pd.Series([0.34, 0.46, 0.62])))


if __name__ == "__main__":
    unittest.main()



import io
from setuptools import find_packages, setup


def get_requirements():
    requirements = open("requirements.txt", "r").read()
    return list(filter(lambda x: x != "", requirements.split()))


main_ns = {}
exec(open("psense_common/version.py").read(), main_ns)

setup(
    name="psense_common",
    version=main_ns["__version__"],
    description="PercuSense Common Modules",
    author="Brad Liang",
    author_email="brad.liang@percusense.com",
    url="https://bitbucket.org/psense/psense-common",
    license="MIT",
    packages=find_packages(),
    long_description=io.open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    install_requires=get_requirements(),
    extras_require={"dev": ["pytest", "flake8", "black"]},
    test_suite="tests",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)
